from __future__ import annotations

import re
from pathlib import Path

import workflows.recipe
from gemmi import cif
from pydantic import BaseModel, Field, ValidationError
from workflows.services.common_service import CommonService

from cryoemservices.util.spa_output_files import get_optics_table
from cryoemservices.util.spa_relion_service_options import RelionServiceOptions


class SelectParticlesParameters(BaseModel):
    input_file: str = Field(..., min_length=1)
    batch_size: int
    image_size: int
    incomplete_batch_size: int = 10000
    relion_options: RelionServiceOptions


class SelectParticles(CommonService):
    """
    A service for batching particles
    """

    # Human readable service name
    _service_name = "SelectParticles"

    # Logger name
    _logger_name = "cryoemservices.services.select_particles"

    # Job name
    job_type = "relion.select.split"

    def initializing(self):
        """Subscribe to a queue. Received messages must be acknowledged."""
        self.log.info("Select particles service starting")
        workflows.recipe.wrap_subscribe(
            self._transport,
            "select_particles",
            self.select_particles,
            acknowledgement=True,
            log_extender=self.extend_log,
            allow_non_recipe_messages=True,
        )

    def select_particles(self, rw, header: dict, message: dict):
        class MockRW:
            def dummy(self, *args, **kwargs):
                pass

        if not rw:
            print(
                "Incoming message is not a recipe message. Simple messages can be valid"
            )
            if (
                not isinstance(message, dict)
                or not message.get("parameters")
                or not message.get("content")
            ):
                self.log.error("Rejected invalid simple message")
                self._transport.nack(header)
                return
            self.log.debug("Received a simple message")

            # Create a wrapper-like object that can be passed to functions
            # as if a recipe wrapper was present.
            rw = MockRW()
            rw.transport = self._transport
            rw.recipe_step = {"parameters": message["parameters"]}
            rw.environment = {"has_recipe_wrapper": False}
            rw.set_default_channel = rw.dummy
            rw.send = rw.dummy
            message = message["content"]

        try:
            if isinstance(message, dict):
                select_params = SelectParticlesParameters(
                    **{**rw.recipe_step.get("parameters", {}), **message}
                )
            else:
                select_params = SelectParticlesParameters(
                    **{**rw.recipe_step.get("parameters", {})}
                )
        except (ValidationError, TypeError) as e:
            self.log.warning(
                f"Selection parameter validation failed for message: {message} "
                f"and recipe parameters: {rw.recipe_step.get('parameters', {})} "
                f"with exception: {e}"
            )
            rw.transport.nack(header)
            return

        self.log.info(f"Inputs: {select_params.input_file}")
        select_params.relion_options.batch_size = select_params.batch_size

        extract_job_dir = Path(
            re.search(".+/job[0-9]{3}/", select_params.input_file)[0]
        )
        project_dir = extract_job_dir.parent.parent

        select_job_num = (
            int(re.search("/job[0-9]{3}", str(extract_job_dir))[0][4:7]) + 1
        )
        select_dir = project_dir / f"Select/job{select_job_num:03}"
        select_dir.mkdir(parents=True, exist_ok=True)

        extracted_parts_file = cif.read_file(select_params.input_file)
        extracted_parts_block = extracted_parts_file.sole_block()
        extracted_parts_loop = extracted_parts_block.find_loop(
            "_rlnCoordinateX"
        ).get_loop()

        current_splits = list(select_dir.glob("particles_split*.star"))
        try:
            num_new_parts = extracted_parts_loop.length()
            num_remaining_parts = extracted_parts_loop.length()
        except AttributeError:
            self.log.info("No particles found for selection")
            num_new_parts = 0
            num_remaining_parts = 0
        if current_splits:
            # If this is a continuation, find the previous split files
            last_split = 1
            for split_file in current_splits:
                split_number = int(re.search("split[0-9]+", str(split_file))[0][5:])
                if split_number > last_split:
                    last_split = split_number
            select_output_file = f"{select_dir}/particles_split{last_split}.star"

            particles_cif = cif.read_file(select_output_file)
            prev_parts_block = particles_cif.find_block("particles")
            prev_parts_loop = prev_parts_block.find_loop("_rlnCoordinateX").get_loop()

            previous_batch_count = prev_parts_loop.length()
            num_prev_parts = prev_parts_loop.length()
            # While we have particles to add and the file is not full
            while num_prev_parts < select_params.batch_size and num_remaining_parts > 0:
                new_row = []
                for col in range(extracted_parts_loop.width()):
                    new_row.append(
                        extracted_parts_loop[num_new_parts - num_remaining_parts, col]
                    )
                prev_parts_loop.add_row(new_row)

                num_prev_parts += 1
                num_remaining_parts -= 1

            particles_cif.write_file(f"{select_output_file}.tmp")
            Path(f"{select_output_file}.tmp").rename(select_output_file)
        else:
            # If this is the first time we ran the job create a new particle split
            # Set this to be split zero so the while loop starts from one
            previous_batch_count = 0
            select_output_file = str(select_dir / "particles_split0.star")

        new_finished_files = []
        # If we filled the last file we need a new one for the remaining particles
        while num_remaining_parts > 0:
            new_split = int(re.search("split[0-9]+", select_output_file)[0][5:]) + 1
            if new_split != 1:
                new_finished_files.append(new_split - 1)
            select_output_file = f"{select_dir}/particles_split{new_split}.star"
            new_particles_cif = get_optics_table(
                select_params.relion_options,
                particle=True,
                im_size=select_params.image_size,
            )

            new_split_block = new_particles_cif.add_new_block("particles")
            new_split_loop = new_split_block.init_loop(
                "_rln",
                [
                    "CoordinateX",
                    "CoordinateY",
                    "ImageName",
                    "MicrographName",
                    "OpticsGroup",
                    "CtfMaxResolution",
                    "CtfFigureOfMerit",
                    "DefocusU",
                    "DefocusV",
                    "DefocusAngle",
                    "CtfBfactor",
                    "CtfScalefactor",
                    "PhaseShift",
                ],
            )

            num_prev_parts = 0
            # While we have particles to add and the file is not full
            while num_prev_parts < select_params.batch_size and num_remaining_parts > 0:
                new_row = []
                for col in range(extracted_parts_loop.width()):
                    new_row.append(
                        extracted_parts_loop[num_new_parts - num_remaining_parts, col]
                    )
                new_split_loop.add_row(new_row)

                num_prev_parts += 1
                num_remaining_parts -= 1

            new_particles_cif.write_file(f"{select_output_file}.tmp")
            Path(f"{select_output_file}.tmp").rename(select_output_file)

        # Send to node creator if a new file was made or there isn't a complete batch
        if (
            select_output_file == f"{select_dir}/particles_split1.star"
            or new_finished_files
        ):
            self.log.info(f"Sending {self.job_type} to node creator")
            node_creator_params = {
                "job_type": self.job_type,
                "input_file": select_params.input_file,
                "output_file": select_output_file,
                "relion_options": dict(select_params.relion_options),
                "command": "",
                "stdout": "",
                "stderr": "",
            }
            if isinstance(rw, MockRW):
                rw.transport.send(
                    destination="node_creator",
                    message={"parameters": node_creator_params, "content": "dummy"},
                )
            else:
                rw.send_to("node_creator", node_creator_params)

        class2d_params = {
            "class2d_dir": f"{project_dir}/Class2D/job",
            "batch_size": select_params.batch_size,
        }
        if select_output_file == f"{select_dir}/particles_split1.star":
            # If still on the first file then register it with murfey
            send_to_2d_classification = False
            if previous_batch_count == 0:
                # First run of this job, do it with any number of particles
                send_to_2d_classification = True
            else:
                # Re-run only if a multiple of the run size is passed
                previous_batch_multiple = (
                    previous_batch_count // select_params.incomplete_batch_size
                )
                new_batch_multiple = (
                    previous_batch_count + num_new_parts
                ) // select_params.incomplete_batch_size
                if new_batch_multiple > previous_batch_multiple:
                    send_to_2d_classification = True

            if send_to_2d_classification:
                self.log.info(
                    f"Sending incomplete batch {select_output_file} to Murfey"
                )
                class2d_params["particles_file"] = f"{select_dir}/particles_split1.star"
                murfey_params = {
                    "register": "incomplete_particles_file",
                    "class2d_message": class2d_params,
                }
                if isinstance(rw, MockRW):
                    rw.transport.send("murfey_feedback", murfey_params)
                else:
                    rw.send_to("murfey_feedback", murfey_params)

        if new_finished_files:
            for new_split in new_finished_files:
                # Set up Class2D job parameters
                class2d_params[
                    "particles_file"
                ] = f"{select_dir}/particles_split{new_split}.star"

                # Send all newly completed files to murfey
                self.log.info(
                    f"Sending complete batch {class2d_params['particles_file']} to Murfey"
                )
                murfey_params = {
                    "register": "complete_particles_file",
                    "class2d_message": class2d_params,
                }
                if isinstance(rw, MockRW):
                    rw.transport.send(
                        destination="murfey_feedback", message=murfey_params
                    )
                else:
                    rw.send_to("murfey_feedback", murfey_params)

        murfey_confirmation = {
            "register": "done_particle_selection",
        }
        if isinstance(rw, MockRW):
            rw.transport.send(
                destination="murfey_feedback", message=murfey_confirmation
            )
        else:
            rw.send_to("murfey_feedback", murfey_confirmation)

        self.log.info(f"Done {self.job_type} for {select_params.input_file}.")
        rw.transport.ack(header)



from __future__ import annotations

import json
import logging
import os
import re
import subprocess
from pathlib import Path
from typing import Optional

import healpy as hp
import matplotlib.pyplot as plt
import numpy as np
from gemmi import cif
from pydantic import BaseModel, Field, ValidationError
from zocalo.wrapper import BaseWrapper

from cryoemservices.util.spa_relion_service_options import (
    RelionServiceOptions,
    update_relion_options,
)

logger = logging.getLogger("cryoemservices.wrappers.class3d")

job_type = "relion.class3d"


class Class3DParameters(BaseModel):
    particles_file: str = Field(..., min_length=1)
    class3d_dir: str = Field(..., min_length=1)
    batch_size: int
    particle_diameter: float = 0
    mask_diameter: float = 190
    do_initial_model: bool = False
    initial_model_file: Optional[str] = None
    initial_model_iterations: int = 200
    initial_model_offset_range: float = 6
    initial_model_offset_step: float = 2
    start_initial_model_C1: bool = True
    dont_combine_weights_via_disc: bool = True
    preread_images: bool = True
    scratch_dir: Optional[str] = None
    nr_pool: int = 10
    pad: int = 2
    skip_gridding: bool = False
    dont_correct_greyscale: bool = True
    initial_lowpass: float = 40.0
    do_ctf: bool = True
    ctf_intact_first_peak: bool = False
    class3d_nr_iter: int = 20
    fast_subsets: bool = False
    tau_fudge: float = 4
    class3d_nr_classes: int = 4
    flatten_solvent: bool = True
    do_zero_mask: bool = True
    highres_limit: Optional[float] = None
    fn_mask: Optional[str] = None
    oversampling: int = 1
    skip_align: bool = False
    healpix_order: int = 2
    offset_range: float = 5
    offset_step: float = 4
    allow_coarser: bool = False
    symmetry: str = "C1"
    do_norm: bool = True
    do_scale: bool = True
    mpi_run_command: str = "srun -n 5"
    threads: int = 8
    gpus: str = "0:1:2:3"
    initial_model_gpus: str = "0,1,2,3"
    picker_id: int
    class3d_grp_uuid: int
    class_uuids: str
    relion_options: RelionServiceOptions


class Class3DWrapper(BaseWrapper):
    """
    A wrapper for the Relion 3D classification job.
    """

    common_flags = {
        "dont_combine_weights_via_disc": "--dont_combine_weights_via_disc",
        "preread_images": "--preread_images",
        "scratch_dir": "--scratch_dir",
        "nr_pool": "--pool",
        "pad": "--pad",
        "skip_gridding": "--skip_gridding",
        "do_ctf": "--ctf",
        "ctf_intact_first_peak": "--ctf_intact_first_peak",
        "class3d_nr_classes": "--K",
        "flatten_solvent": "--flatten_solvent",
        "do_zero_mask": "--zero_mask",
        "oversampling": "--oversampling",
        "healpix_order": "--healpix_order",
        "threads": "--j",
    }

    # Values for ISPyB lookups
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.class_uuids_dict = {}
        self.class_uuids_keys = []

    def run_initial_model(self, initial_model_params, project_dir, job_num):
        """
        Run the initial model for 3D classification and register results
        """
        job_dir = project_dir / f"InitialModel/job{job_num:03}/"
        job_dir.mkdir(parents=True, exist_ok=True)
        particles_file = str(
            Path(initial_model_params.particles_file).relative_to(project_dir)
        )

        initial_model_flags = {
            "initial_model_iterations": "--iter",
            "initial_model_offset_range": "--offset_range",
            "initial_model_offset_step": "--offset_step",
        }
        initial_model_flags.update(self.common_flags)

        initial_model_command = [
            "relion_refine",
            "--grad",
            "--denovo_3dref",
            "--i",
            particles_file,
            "--o",
            f"{job_dir.relative_to(project_dir)}/run",
            "--particle_diameter",
            f"{initial_model_params.relion_options.mask_diameter}",
            "--gpu",
            initial_model_params.initial_model_gpus,
        ]
        if initial_model_params.start_initial_model_C1:
            initial_model_command.extend(("--sym", "C1"))
        else:
            initial_model_command.extend(("--sym", initial_model_params.symmetry))
        for k, v in initial_model_params.dict().items():
            if v and (k in initial_model_flags):
                if type(v) is bool:
                    initial_model_command.append(initial_model_flags[k])
                else:
                    initial_model_command.extend((initial_model_flags[k], str(v)))
        initial_model_command.extend(
            ("--pipeline_control", f"{job_dir.relative_to(project_dir)}/")
        )

        # Run initial model and confirm it ran successfully
        self.log.info("Running initial model")
        result = subprocess.run(
            initial_model_command, cwd=str(project_dir), capture_output=True
        )

        # Register the initial model job with the node creator
        self.log.info("Sending relion.initialmodel (model) to node creator")
        node_creator_parameters = {
            "job_type": "relion.initialmodel",
            "input_file": f"{project_dir}/{particles_file}",
            "output_file": f"{job_dir}/initial_model.mrc",
            "relion_options": dict(initial_model_params.relion_options),
            "command": " ".join(initial_model_command),
            "stdout": result.stdout.decode("utf8", "replace"),
            "stderr": result.stderr.decode("utf8", "replace"),
        }
        if result.returncode:
            node_creator_parameters["success"] = False
        else:
            node_creator_parameters["success"] = True
        self.recwrap.send_to("node_creator", node_creator_parameters)

        # End here if the command failed
        if result.returncode:
            self.log.error(
                f"Relion initial model failed with exitcode {result.returncode}:\n"
                + result.stderr.decode("utf8", "replace")
            )
            return False

        ini_model_file = job_dir / "initial_model.mrc"
        align_symmetry_command = [
            "relion_align_symmetry",
            "--i",
            str(
                job_dir.relative_to(project_dir)
                / f"run_it{initial_model_params.initial_model_iterations:03}_model.star"
            ),
            "--o",
            f"{ini_model_file.relative_to(project_dir)}",
            "--sym",
            initial_model_params.symmetry,
            "--apply_sym",
            "--select_largest_class",
            "--pipeline_control",
            f"{job_dir.relative_to(project_dir)}/",
        ]

        # Run symmetry alignment and confirm it ran successfully
        self.log.info("Running symmetry alignment")
        result = subprocess.run(
            align_symmetry_command, cwd=str(project_dir), capture_output=True
        )

        # Register the initial model job with the node creator
        self.log.info("Sending relion.initialmodel (alignment) to node creator")
        node_creator_parameters = {
            "job_type": "relion.initialmodel",
            "input_file": f"{project_dir}/{particles_file}",
            "output_file": f"{job_dir}/initial_model.mrc",
            "relion_options": dict(initial_model_params.relion_options),
            "command": "".join(align_symmetry_command),
            "stdout": result.stdout.decode("utf8", "replace"),
            "stderr": result.stderr.decode("utf8", "replace"),
        }
        if result.returncode:
            node_creator_parameters["success"] = False
        else:
            node_creator_parameters["success"] = True
        self.recwrap.send_to("node_creator", node_creator_parameters)

        # End here if the command failed
        if result.returncode:
            self.log.error(
                f"Relion initial model symmetry alignment "
                f"failed with exitcode {result.returncode}:\n"
                + result.stderr.decode("utf8", "replace")
            )
            return "", {}

        # Send Murfey the location of the initial model
        murfey_params = {
            "register": "save_initial_model",
            "initial_model": f"{job_dir}/initial_model.mrc",
        }
        self.recwrap.send_to("murfey_feedback", murfey_params)

        # Extract parameters for ispyb
        initial_model_cif = cif.read_file(
            f"{job_dir}/run_it{initial_model_params.initial_model_iterations:03}_model.star"
        )
        initial_model_block = initial_model_cif.find_block("model_classes")
        model_scores = list(initial_model_block.find_loop("_rlnClassDistribution"))
        model_resolutions = list(
            initial_model_block.find_loop("_rlnEstimatedResolution")
        )
        resolution = model_resolutions[model_scores == max(model_scores)]
        number_of_particles = (
            float(model_scores[model_scores == max(model_scores)])
            * initial_model_params.batch_size
        )

        self.log.info(
            "Will send best initial model to ispyb with "
            f"resolution {resolution} and {number_of_particles} particles"
        )
        ini_ispyb_parameters = [
            {
                "ispyb_command": "buffer",
                "buffer_lookup": {
                    "particle_classification_id": self.class_uuids_dict[
                        self.class_uuids_keys[0]
                    ]
                },
                "buffer_command": {"ispyb_command": "insert_cryoem_initial_model"},
                "number_of_particles": number_of_particles,
                "resolution": resolution,
                "store_result": "ispyb_initial_model_id",
            }
        ]
        for i in range(1, initial_model_params.class3d_nr_classes):
            # Insert initial model for every class, sending model id each time
            ini_ispyb_parameters.append(
                {
                    "ispyb_command": "buffer",
                    "buffer_lookup": {
                        "particle_classification_id": self.class_uuids_dict[
                            self.class_uuids_keys[i]
                        ]
                    },
                    "buffer_command": {"ispyb_command": "insert_cryoem_initial_model"},
                    "number_of_particles": number_of_particles,
                    "resolution": resolution,
                    "cryoem_initial_model_id": "$ispyb_initial_model_id",
                }
            )
        for i in range(initial_model_params.class3d_nr_classes):
            # Add resolution to every model if it is finite
            if np.isfinite(float(resolution)):
                ini_ispyb_parameters[i]["resolution"] = resolution
            else:
                ini_ispyb_parameters[i]["resolution"] = 0.0

        self.log.info("Running 3D classification using new initial model")
        return f"{ini_model_file}", ini_ispyb_parameters

    def run(self):
        """
        Run the 3D classification and register results
        """
        assert hasattr(self, "recwrap"), "No recipewrapper object found"
        params_dict = self.recwrap.recipe_step["job_parameters"]
        try:
            class3d_params = Class3DParameters(**params_dict)
        except (ValidationError, TypeError) as e:
            self.log.warning(
                f"Class3D parameter validation failed for parameters: {params_dict} "
                f"with exception: {e}"
            )
            return False

        # Class ids get fed in as a string, need to convert these to a dictionary
        self.class_uuids_dict = json.loads(class3d_params.class_uuids.replace("'", '"'))
        self.class_uuids_keys = list(self.class_uuids_dict.keys())

        # Update the relion options to get out the mask diameter
        class3d_params.relion_options = update_relion_options(
            class3d_params.relion_options, dict(class3d_params)
        )

        # Make the job directory and move to the project directory
        job_dir = Path(class3d_params.class3d_dir)
        if (job_dir / "RELION_JOB_EXIT_SUCCESS").exists():
            # This job over-writes a previous one
            job_is_rerun = True
        else:
            job_is_rerun = False
            job_dir.mkdir(parents=True, exist_ok=True)
        project_dir = job_dir.parent.parent
        os.chdir(project_dir)

        particles_file = str(
            Path(class3d_params.particles_file).relative_to(project_dir)
        )
        self.log.info(f"Running Class3D for {particles_file}")

        # Run the initial model if requested, otherwise look for a pre-existing file
        if class3d_params.do_initial_model:
            job_num_3d = int(
                re.search("/job[0-9]{3}", class3d_params.class3d_dir)[0][4:7]
            )
            initial_model_file, initial_model_ispyb_parameters = self.run_initial_model(
                class3d_params, project_dir, job_num_3d - 1
            )
        else:
            initial_model_file = class3d_params.initial_model_file
            initial_model_ispyb_parameters = {}
        if not initial_model_file:
            # If there isn't an initial model file something has gone wrong
            self.log.error("No initial model file found, stopping.")
            return False

        class3d_flags = {
            "dont_correct_greyscale": "--firstiter_cc",
            "initial_lowpass": "--ini_high",
            "class3d_nr_iter": "--iter",
            "fast_subsets": "--fast_subsets",
            "tau_fudge": "--tau2_fudge",
            "highres_limit": "--strict_highres_exp",
            "fn_mask": "--solvent_mask",
            "skip_align": "--skip_align",
            "offset_range": "--offset_range",
            "offset_step": "--offset_step",
            "allow_coarser": "--allow_coarser_sampling",
            "symmetry": "--sym",
            "do_norm": "--norm",
            "do_scale": "--scale",
            "gpus": "--gpu",
        }
        class3d_flags.update(self.common_flags)

        # Create the classification command
        class3d_command = class3d_params.mpi_run_command.split()
        class3d_command.extend(
            [
                "relion_refine_mpi",
                "--i",
                particles_file,
                "--o",
                f"{job_dir.relative_to(project_dir)}/run",
                "--ref",
                initial_model_file,
                "--particle_diameter",
                f"{class3d_params.relion_options.mask_diameter}",
            ]
        )
        for k, v in class3d_params.dict().items():
            if v and (k in class3d_flags):
                if type(v) is bool:
                    class3d_command.append(class3d_flags[k])
                else:
                    class3d_command.extend((class3d_flags[k], str(v)))
        class3d_command.extend(
            ("--pipeline_control", f"{job_dir.relative_to(project_dir)}/")
        )

        # Run Class3D and confirm it ran successfully
        self.log.info(" ".join(class3d_command))
        result = subprocess.run(
            class3d_command, cwd=str(project_dir), capture_output=True
        )
        if not job_is_rerun:
            (job_dir / "RELION_JOB_EXIT_SUCCESS").unlink(missing_ok=True)

        # Register the Class3D job with the node creator
        self.log.info(f"Sending {job_type} to node creator")
        node_creator_parameters = {
            "job_type": job_type,
            "input_file": class3d_params.particles_file + f":{initial_model_file}",
            "output_file": class3d_params.class3d_dir,
            "relion_options": dict(class3d_params.relion_options),
            "command": " ".join(class3d_command),
            "stdout": result.stdout.decode("utf8", "replace"),
            "stderr": result.stderr.decode("utf8", "replace"),
        }
        if result.returncode:
            node_creator_parameters["success"] = False
        else:
            node_creator_parameters["success"] = True
        self.recwrap.send_to("node_creator", node_creator_parameters)

        # End here if the command failed
        if result.returncode:
            self.log.error(
                f"Relion Class3D failed with exitcode {result.returncode}:\n"
                + result.stderr.decode("utf8", "replace")
            )
            return False

        # Generate healpix image of the particle distribution
        self.log.info("Generating healpix angular distribution image")
        data = cif.read_file(
            str(job_dir / f"run_it{class3d_params.class3d_nr_iter:03}_data.star")
        )
        particles_block = data.find_block("particles")
        if particles_block:
            angles_rot = np.array(
                particles_block.find_loop("_rlnAngleRot"), dtype=float
            )
            angles_tilt = np.array(
                particles_block.find_loop("_rlnAngleTilt"), dtype=float
            )
            class_numbers = np.array(
                particles_block.find_loop("_rlnClassNumber"), dtype=int
            )

            for class_id in range(class3d_params.class3d_nr_classes):
                if not len(angles_tilt[class_numbers == class_id + 1]):
                    # Skip any classes with no particles
                    continue
                try:
                    # Extract counts of particles in each healpix bin
                    angle_pixel_bins = hp.pixelfunc.ang2pix(
                        np.power(2, class3d_params.healpix_order + 1),
                        angles_tilt[class_numbers == class_id + 1] * np.pi / 180,
                        angles_rot[class_numbers == class_id + 1] * np.pi / 180,
                    )
                    bin_ids, pixel_counts = np.unique(
                        angle_pixel_bins, return_counts=True
                    )
                    all_pixel_bins = np.zeros(
                        hp.nside2npix(np.power(2, class3d_params.healpix_order + 1))
                    )
                    all_pixel_bins[bin_ids] = pixel_counts

                    # Create and save the healpix image
                    hp.mollview(
                        all_pixel_bins, title=f"Class {class_id+1} angular distribution"
                    )
                    hp.graticule()
                    plt.savefig(
                        job_dir
                        / f"run_it{class3d_params.class3d_nr_iter:03}_class{class_id+1:03}_angdist.jpeg"
                    )
                except ValueError as e:
                    self.log.warning(f"Healpix failed with error {e}")

        # Send classification job information to ispyb
        ispyb_parameters = []
        classification_grp_ispyb_parameters = {
            "ispyb_command": "buffer",
            "buffer_command": {"ispyb_command": "insert_particle_classification_group"},
            "type": "3D",
            "batch_number": "1",
            "number_of_particles_per_batch": class3d_params.batch_size,
            "number_of_classes_per_batch": class3d_params.class3d_nr_classes,
            "symmetry": class3d_params.symmetry,
            "particle_picker_id": class3d_params.picker_id,
        }
        if job_is_rerun:
            # If this job overwrites another get the id for it
            classification_grp_ispyb_parameters["buffer_lookup"] = {
                "particle_classification_group_id": class3d_params.class3d_grp_uuid,
            }
        else:
            classification_grp_ispyb_parameters[
                "buffer_store"
            ] = class3d_params.class3d_grp_uuid
        ispyb_parameters.append(classification_grp_ispyb_parameters)

        # Send individual classes to ispyb
        class_star_file = cif.read_file(
            f"{class3d_params.class3d_dir}/run_it{class3d_params.class3d_nr_iter:03}_model.star"
        )
        classes_block = class_star_file.find_block("model_classes")
        classes_loop = classes_block.find_loop("_rlnReferenceImage").get_loop()

        for class_id in range(class3d_params.class3d_nr_classes):
            # Add an ispyb insert for each class
            class_ispyb_parameters = {
                "ispyb_command": "buffer",
                "buffer_lookup": {
                    "particle_classification_group_id": class3d_params.class3d_grp_uuid
                },
                "buffer_command": {"ispyb_command": "insert_particle_classification"},
                "class_number": class_id + 1,
                "class_image_full_path": (
                    f"{class3d_params.class3d_dir}/"
                    f"run_it{class3d_params.class3d_nr_iter:03}_class{class_id+1:03}.mrc"
                ),
                "particles_per_class": (
                    float(classes_loop[class_id, 1]) * class3d_params.batch_size
                ),
                "class_distribution": classes_loop[class_id, 1],
                "rotation_accuracy": classes_loop[class_id, 2],
                "translation_accuracy": classes_loop[class_id, 3],
            }
            if job_is_rerun:
                class_ispyb_parameters["buffer_lookup"].update(
                    {
                        "particle_classification_id": self.class_uuids_dict[
                            self.class_uuids_keys[class_id]
                        ],
                    }
                )
            else:
                class_ispyb_parameters["buffer_store"] = self.class_uuids_dict[
                    self.class_uuids_keys[class_id]
                ]

            # Add the resolution and fourier completeness if they are valid numbers
            estimated_resolution = float(classes_loop[class_id, 4])
            if np.isfinite(estimated_resolution):
                class_ispyb_parameters["estimated_resolution"] = estimated_resolution
            else:
                class_ispyb_parameters["estimated_resolution"] = 0.0
            fourier_completeness = float(classes_loop[class_id, 5])
            if np.isfinite(fourier_completeness):
                class_ispyb_parameters[
                    "overall_fourier_completeness"
                ] = fourier_completeness
            else:
                class_ispyb_parameters["overall_fourier_completeness"] = 0.0

            # Add the ispyb command to the command list
            ispyb_parameters.append(class_ispyb_parameters)

        # Add on the initial model insert before sending
        if class3d_params.do_initial_model:
            ispyb_parameters.extend(initial_model_ispyb_parameters)

        self.log.info(f"Sending to ispyb {ispyb_parameters}")
        self.recwrap.send_to(
            "ispyb_connector",
            {
                "ispyb_command": "multipart_message",
                "ispyb_command_list": ispyb_parameters,
            },
        )

        # Tell Murfey the batch has finished
        murfey_params = {
            "register": "done_3d_batch",
        }
        self.recwrap.send_to("murfey_feedback", murfey_params)

        (job_dir / "RELION_JOB_EXIT_SUCCESS").touch(exist_ok=True)
        self.log.info(f"Done {job_type} for {class3d_params.particles_file}.")
        return True



from __future__ import annotations

import json
import logging
import os
import re
import subprocess
from pathlib import Path

import numpy as np
from gemmi import cif
from pydantic import BaseModel, Field, ValidationError
from zocalo.wrapper import BaseWrapper

from cryoemservices.util.spa_relion_service_options import (
    RelionServiceOptions,
    update_relion_options,
)

logger = logging.getLogger("cryoemservices.wrappers.class2d")


class Class2DParameters(BaseModel):
    particles_file: str = Field(..., min_length=1)
    class2d_dir: str = Field(..., min_length=1)
    batch_is_complete: bool
    batch_size: int
    particle_diameter: float = 0
    mask_diameter: float = 190
    do_vdam = False
    dont_combine_weights_via_disc: bool = True
    preread_images: bool = True
    scratch_dir: str = None
    nr_pool: int = 10
    pad: int = 2
    skip_gridding: bool = False
    do_ctf: bool = True
    ctf_intact_first_peak: bool = False
    class2d_nr_iter: int = 20
    tau_fudge: float = 2
    class2d_nr_classes: int = 50
    flatten_solvent: bool = True
    do_zero_mask: bool = True
    highres_limit: float = None
    centre_classes: bool = True
    oversampling: int = 1
    skip_align: bool = False
    psi_step: float = 12.0
    offset_range: float = 5
    offset_step: float = 2
    allow_coarser: bool = False
    do_norm: bool = True
    do_scale: bool = True
    mpi_run_command: str = "srun -n 5"
    threads: int = 8
    gpus: str = "0:1:2:3"
    picker_id: int
    class2d_grp_uuid: int
    class_uuids: str
    do_icebreaker_jobs: bool = True
    relion_options: RelionServiceOptions


class Class2DWrapper(BaseWrapper):
    """
    A wrapper for the Relion 2D classification job.
    """

    # Values to extract for ISPyB
    previous_total_count = 0
    total_count = 0

    # Values for ISPyB lookups
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.class_uuids_dict = {}
        self.class_uuids_keys = []

    def parse_combiner_output(self, combiner_stdout: str):
        """
        Read the output logs of the star file combination
        """
        for line in combiner_stdout.split("\n"):
            if line.startswith("Adding") and "particles_all.star" in line:
                line_split = line.split()
                self.previous_total_count = int(line_split[3])

            if line.startswith("Combined"):
                line_split = line.split()
                self.total_count = int(line_split[6])

    def run(self):
        """
        Run the 2D classification and register results
        """
        assert hasattr(self, "recwrap"), "No recipewrapper object found"
        params_dict = self.recwrap.recipe_step["job_parameters"]
        try:
            class2d_params = Class2DParameters(**params_dict)
        except (ValidationError, TypeError) as e:
            self.log.warning(
                f"Class2D parameter validation failed for parameters: {params_dict} "
                f"with exception: {e}"
            )
            return False

        # Class ids get fed in as a string, need to convert these to a dictionary
        self.class_uuids_dict = json.loads(class2d_params.class_uuids.replace("'", '"'))
        self.class_uuids_keys = list(self.class_uuids_dict.keys())

        if class2d_params.do_vdam:
            job_type = "relion.class2d.vdam"
        else:
            job_type = "relion.class2d.em"

        # Update the relion options to get out the mask diameter
        class2d_params.relion_options = update_relion_options(
            class2d_params.relion_options, dict(class2d_params)
        )

        # Make the job directory and move to the project directory
        job_dir = Path(class2d_params.class2d_dir)
        if (job_dir / "RELION_JOB_EXIT_SUCCESS").exists():
            # This job over-writes a previous one
            job_is_rerun = True
        else:
            job_is_rerun = False
            job_dir.mkdir(parents=True, exist_ok=True)
        project_dir = job_dir.parent.parent
        os.chdir(project_dir)
        job_num = int(
            re.search("/job[0-9]{3}", str(class2d_params.class2d_dir))[0][4:7]
        )

        particles_file = str(
            Path(class2d_params.particles_file).relative_to(project_dir)
        )
        self.log.info(f"Running Class2D for {particles_file}")

        class2d_flags = {
            "dont_combine_weights_via_disc": "--dont_combine_weights_via_disc",
            "preread_images": "--preread_images",
            "scratch_dir": "--scratch_dir",
            "nr_pool": "--pool",
            "pad": "--pad",
            "skip_gridding": "--skip_gridding",
            "do_ctf": "--ctf",
            "ctf_intact_first_peak": "--ctf_intact_first_peak",
            "class2d_nr_iter": "--iter",
            "tau_fudge": "--tau2_fudge",
            "class2d_nr_classes": "--K",
            "flatten_solvent": "--flatten_solvent",
            "do_zero_mask": "--zero_mask",
            "highres_limit": "--strict_highres_exp",
            "centre_classes": "--center_classes",
            "oversampling": "--oversampling",
            "skip_align": "--skip_align",
            "psi_step": "--psi_step",
            "offset_range": "--offset_range",
            "offset_step": "--offset_step",
            "allow_coarser": "--allow_coarser_sampling",
            "do_norm": "--norm",
            "do_scale": "--scale",
            "threads": "--j",
            "gpus": "--gpu",
        }

        # Create the classification command
        class2d_command = class2d_params.mpi_run_command.split()
        class2d_command.extend(
            [
                "relion_refine_mpi",
                "--i",
                particles_file,
                "--o",
                f"{job_dir.relative_to(project_dir)}/run",
                "--particle_diameter",
                f"{class2d_params.relion_options.mask_diameter}",
            ]
        )
        for k, v in class2d_params.dict().items():
            if v and (k in class2d_flags):
                if type(v) is bool:
                    class2d_command.append(class2d_flags[k])
                else:
                    class2d_command.extend((class2d_flags[k], str(v)))
        class2d_command.extend(
            ("--pipeline_control", f"{job_dir.relative_to(project_dir)}/")
        )

        # Run Class2D and confirm it ran successfully
        self.log.info(" ".join(class2d_command))
        result = subprocess.run(
            class2d_command, cwd=str(project_dir), capture_output=True
        )
        if not job_is_rerun:
            (job_dir / "RELION_JOB_EXIT_SUCCESS").unlink(missing_ok=True)

        # Register the Class2D job with the node creator
        self.log.info(f"Sending {job_type} to node creator")
        node_creator_parameters = {
            "job_type": job_type,
            "input_file": class2d_params.particles_file,
            "output_file": class2d_params.class2d_dir,
            "relion_options": dict(class2d_params.relion_options),
            "command": " ".join(class2d_command),
            "stdout": result.stdout.decode("utf8", "replace"),
            "stderr": result.stderr.decode("utf8", "replace"),
        }
        if result.returncode:
            node_creator_parameters["success"] = False
        else:
            node_creator_parameters["success"] = True
        self.recwrap.send_to("node_creator", node_creator_parameters)

        # End here if the command failed
        if result.returncode:
            self.log.error(
                f"Relion Class2D failed with exitcode {result.returncode}:\n"
                + result.stderr.decode("utf8", "replace")
            )
            # Tell Murfey to release the hold on 2D classification jobs
            murfey_params = {
                "register": "done_incomplete_2d_batch",
                "job_dir": class2d_params.class2d_dir,
            }
            self.recwrap.send_to("murfey_feedback", murfey_params)
            return False

        # Send classification job information to ispyb
        if not class2d_params.batch_is_complete:
            class_particles_file = cif.read_file(
                f"{class2d_params.class2d_dir}/run_it{class2d_params.class2d_nr_iter:03}_data.star"
            )
            particles_block = class_particles_file.find_block("particles")
            particles_in_batch = len(particles_block.find_loop("_rlnCoordinateX"))
        else:
            particles_in_batch = class2d_params.batch_size

        ispyb_parameters = []
        classification_grp_ispyb_parameters = {
            "ispyb_command": "buffer",
            "buffer_command": {"ispyb_command": "insert_particle_classification_group"},
            "type": "2D",
            "batch_number": int(
                class2d_params.particles_file.split("particles_split")[1].split(".")[0]
            ),
            "number_of_particles_per_batch": particles_in_batch,
            "number_of_classes_per_batch": class2d_params.class2d_nr_classes,
            "symmetry": "C1",
            "particle_picker_id": class2d_params.picker_id,
        }
        if job_is_rerun:
            # If this job overwrites another get the id for it
            classification_grp_ispyb_parameters["buffer_lookup"] = {
                "particle_classification_group_id": class2d_params.class2d_grp_uuid,
            }
        else:
            classification_grp_ispyb_parameters[
                "buffer_store"
            ] = class2d_params.class2d_grp_uuid
        ispyb_parameters.append(classification_grp_ispyb_parameters)

        # Send individual classes to ispyb
        class_star_file = cif.read_file(
            f"{class2d_params.class2d_dir}/run_it{class2d_params.class2d_nr_iter:03}_model.star"
        )
        classes_block = class_star_file.find_block("model_classes")
        classes_loop = classes_block.find_loop("_rlnReferenceImage").get_loop()

        for class_id in range(class2d_params.class2d_nr_classes):
            # Add an ispyb insert for each class
            class_ispyb_parameters = {
                "ispyb_command": "buffer",
                "buffer_lookup": {
                    "particle_classification_group_id": class2d_params.class2d_grp_uuid
                },
                "buffer_command": {"ispyb_command": "insert_particle_classification"},
                "class_number": class_id + 1,
                "class_image_full_path": (
                    f"{class2d_params.class2d_dir}"
                    f"/run_it{class2d_params.class2d_nr_iter:03}_classes_{class_id+1}.jpeg"
                ),
                "particles_per_class": (
                    float(classes_loop[class_id, 1]) * particles_in_batch
                ),
                "class_distribution": classes_loop[class_id, 1],
                "rotation_accuracy": classes_loop[class_id, 2],
                "translation_accuracy": classes_loop[class_id, 3],
            }
            if job_is_rerun:
                class_ispyb_parameters["buffer_lookup"].update(
                    {
                        "particle_classification_id": self.class_uuids_dict[
                            self.class_uuids_keys[class_id]
                        ]
                    }
                )
            else:
                class_ispyb_parameters["buffer_store"] = self.class_uuids_dict[
                    self.class_uuids_keys[class_id]
                ]

            # Add the resolution and fourier completeness if they are valid numbers
            estimated_resolution = float(classes_loop[class_id, 4])
            if np.isfinite(estimated_resolution):
                class_ispyb_parameters["estimated_resolution"] = estimated_resolution
            else:
                class_ispyb_parameters["estimated_resolution"] = 0.0
            fourier_completeness = float(classes_loop[class_id, 5])
            if np.isfinite(fourier_completeness):
                class_ispyb_parameters[
                    "overall_fourier_completeness"
                ] = fourier_completeness
            else:
                class_ispyb_parameters["overall_fourier_completeness"] = 0.0

            # Add the ispyb command to the command list
            ispyb_parameters.append(class_ispyb_parameters)

        # Send a request to make the class images
        self.log.info("Sending to images service")
        self.recwrap.send_to(
            "images",
            {
                "image_command": "mrc_to_jpeg",
                "file": (
                    f"{class2d_params.class2d_dir}"
                    f"/run_it{class2d_params.class2d_nr_iter:03}_classes.mrcs"
                ),
                "all_frames": "True",
            },
        )

        # Send all the ispyb class insertion commands
        self.log.info(f"Sending to ispyb {ispyb_parameters}")
        self.recwrap.send_to(
            "ispyb_connector",
            {
                "ispyb_command": "multipart_message",
                "ispyb_command_list": ispyb_parameters,
            },
        )

        if class2d_params.batch_is_complete:
            # Create an icebreaker job
            if class2d_params.do_icebreaker_jobs:
                self.log.info("Sending to icebreaker particle analysis")
                icebreaker_params = {
                    "icebreaker_type": "particles",
                    "input_micrographs": (
                        f"{project_dir}/IceBreaker/job003/grouped_micrographs.star"
                    ),
                    "input_particles": class2d_params.particles_file,
                    "output_path": f"{project_dir}/IceBreaker/job{job_num + 1:03}/",
                    "mc_uuid": -1,
                    "relion_options": dict(class2d_params.relion_options),
                }
                self.recwrap.send_to("icebreaker", icebreaker_params)

            # Create a 2D autoselection job
            self.log.info("Sending to class selection")
            autoselect_parameters = {
                "input_file": f"{class2d_params.class2d_dir}/run_it{class2d_params.class2d_nr_iter:03}_optimiser.star",
                "relion_options": dict(class2d_params.relion_options),
                "class_uuids": class2d_params.class_uuids,
            }
            self.recwrap.send_to("select_classes", autoselect_parameters)
        else:
            # Tell Murfey the incomplete batch has finished
            murfey_params = {
                "register": "done_incomplete_2d_batch",
                "job_dir": class2d_params.class2d_dir,
            }
            self.recwrap.send_to("murfey_feedback", murfey_params)

        (job_dir / "RELION_JOB_EXIT_SUCCESS").touch(exist_ok=True)
        self.log.info(f"Done {job_type} for {class2d_params.particles_file}.")
        return True



from __future__ import annotations

import io
import json
import re
import shutil
import subprocess
from contextlib import redirect_stdout
from pathlib import Path

import numpy as np
import workflows.recipe
from gemmi import cif
from pydantic import BaseModel, Field, ValidationError
from workflows.services.common_service import CommonService

from cryoemservices.pipeliner_plugins.combine_star_files import (
    combine_star_files,
    split_star_file,
)
from cryoemservices.util.spa_relion_service_options import RelionServiceOptions


class SelectClassesParameters(BaseModel):
    input_file: str = Field(..., min_length=1)
    combine_star_job_number: int
    class2d_fraction_of_classes_to_remove: float = 0.9
    particles_file: str = "particles.star"
    classes_file: str = "class_averages.star"
    python_exe: str = "python"
    autoselect_min_score: float = 0
    min_particles: int = 500
    class3d_batch_size: int = 50000
    class3d_max_size: int = 200000
    class_uuids: str
    relion_options: RelionServiceOptions


class SelectClasses(CommonService):
    """
    A service for running Relion autoselection on 2D classes
    """

    # Human readable service name
    _service_name = "SelectClasses"

    # Logger name
    _logger_name = "cryoemservices.services.select_classes"

    # Job name
    job_type = "relion.select.class2dauto"

    # Values to extract for ISPyB
    previous_total_count = 0
    total_count = 0

    # Values for ISPyB lookups
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.class_uuids_dict = {}
        self.class_uuids_keys = []

    def initializing(self):
        """Subscribe to a queue. Received messages must be acknowledged."""
        self.log.info("Select classes service starting")
        workflows.recipe.wrap_subscribe(
            self._transport,
            "select_classes",
            self.select_classes,
            acknowledgement=True,
            log_extender=self.extend_log,
            allow_non_recipe_messages=True,
        )

    def parse_combiner_output(self, combiner_stdout: str):
        """
        Read the output logs of the star file combination
        """
        for line in combiner_stdout.split("\n"):
            if line.startswith("Adding") and "particles_all.star" in line:
                line_split = line.split()
                self.previous_total_count = int(line_split[3])

            if line.startswith("Combined"):
                line_split = line.split()
                self.total_count = int(line_split[6])

    def select_classes(self, rw, header: dict, message: dict):
        class MockRW:
            def dummy(self, *args, **kwargs):
                pass

        if not rw:
            print(
                "Incoming message is not a recipe message. Simple messages can be valid"
            )
            if (
                not isinstance(message, dict)
                or not message.get("parameters")
                or not message.get("content")
            ):
                self.log.error("Rejected invalid simple message")
                self._transport.nack(header)
                return
            self.log.debug("Received a simple message")

            # Create a wrapper-like object that can be passed to functions
            # as if a recipe wrapper was present.
            rw = MockRW()
            rw.transport = self._transport
            rw.recipe_step = {"parameters": message["parameters"]}
            rw.environment = {"has_recipe_wrapper": False}
            rw.set_default_channel = rw.dummy
            rw.send = rw.dummy
            message = message["content"]

        try:
            if isinstance(message, dict):
                autoselect_params = SelectClassesParameters(
                    **{**rw.recipe_step.get("parameters", {}), **message}
                )
            else:
                autoselect_params = SelectClassesParameters(
                    **{**rw.recipe_step.get("parameters", {})}
                )
        except (ValidationError, TypeError) as e:
            self.log.warning(
                f"Selection parameter validation failed for message: {message} "
                f"and recipe parameters: {rw.recipe_step.get('parameters', {})} "
                f"with exception: {e}"
            )
            rw.transport.nack(header)
            return

        # Update the relion options
        autoselect_params.relion_options.class2d_fraction_of_classes_to_remove = (
            autoselect_params.class2d_fraction_of_classes_to_remove
        )
        autoselect_params.relion_options.autoselect_min_score = (
            autoselect_params.autoselect_min_score
        )

        self.log.info(f"Inputs: {autoselect_params.input_file}")

        class2d_job_dir = Path(
            re.search(".+/job[0-9]{3}/", autoselect_params.input_file)[0]
        )
        project_dir = class2d_job_dir.parent.parent

        select_job_num = (
            int(re.search("/job[0-9]{3}", str(class2d_job_dir))[0][4:7]) + 2
        )
        select_dir = project_dir / f"Select/job{select_job_num:03}"
        select_dir.mkdir(parents=True, exist_ok=True)

        autoselect_flags = {
            "particles_file": "--fn_sel_parts",
            "classes_file": "--fn_sel_classavgs",
            "python_exe": "--python",
            "min_particles": "--select_min_nr_particles",
        }
        # Create the class selection command
        autoselect_command = [
            "relion_class_ranker",
            "--opt",
            autoselect_params.input_file,
            "--o",
            f"{select_dir.relative_to(project_dir)}/",
            "--auto_select",
            "--fn_root",
            "rank",
            "--do_granularity_features",
        ]
        for k, v in autoselect_params.dict().items():
            if v and (k in autoselect_flags):
                autoselect_command.extend((autoselect_flags[k], str(v)))
        autoselect_command.extend(
            ("--pipeline_control", f"{select_dir.relative_to(project_dir)}/")
        )

        if not autoselect_params.autoselect_min_score:
            autoselect_command.extend(("--min_score", "0.0"))
        else:
            autoselect_command.extend(
                ("--min_score", str(autoselect_params.autoselect_min_score))
            )

        # Run the class selection
        autoselect_result = subprocess.run(
            autoselect_command, cwd=str(project_dir), capture_output=True
        )

        if (
            not autoselect_params.autoselect_min_score
            and not autoselect_result.returncode
        ):
            # If a minimum score isn't given, then work it out and rerun the job
            star_doc = cif.read_file(str(select_dir / "rank_model.star"))
            star_block = star_doc["model_classes"]
            class_scores = np.array(star_block.find_loop("_rlnClassScore"), dtype=float)
            quantile_threshold = np.quantile(
                class_scores,
                float(autoselect_params.class2d_fraction_of_classes_to_remove),
            )

            self.log.info(f"Sending new threshold {quantile_threshold} to Murfey")
            murfey_params = {
                "register": "save_class_selection_score",
                "class_selection_score": quantile_threshold,
            }
            if isinstance(rw, MockRW):
                rw.transport.send(destination="murfey_feedback", message=murfey_params)
            else:
                rw.send_to("murfey_feedback", murfey_params)

            self.log.info(
                f"Re-running class selection with new threshold {quantile_threshold}"
            )
            autoselect_command[-1] = str(quantile_threshold)

            # Re-run the class selection
            autoselect_result = subprocess.run(
                autoselect_command, cwd=str(project_dir), capture_output=True
            )

        # Send class selection job to node creator
        self.log.info(f"Sending {self.job_type} to node creator")
        autoselect_node_creator_params = {
            "job_type": self.job_type,
            "input_file": autoselect_params.input_file,
            "output_file": str(select_dir / autoselect_params.particles_file),
            "relion_options": dict(autoselect_params.relion_options),
            "command": " ".join(autoselect_command),
            "stdout": autoselect_result.stdout.decode("utf8", "replace"),
            "stderr": autoselect_result.stderr.decode("utf8", "replace"),
        }
        if autoselect_result.returncode:
            autoselect_node_creator_params["success"] = False
        else:
            autoselect_node_creator_params["success"] = True
        if isinstance(rw, MockRW):
            rw.transport.send(
                destination="node_creator",
                message={
                    "parameters": autoselect_node_creator_params,
                    "content": "dummy",
                },
            )
        else:
            rw.send_to("node_creator", autoselect_node_creator_params)

        # End here if the command failed
        if autoselect_result.returncode:
            self.log.error(
                f"2D autoselection failed with exitcode {autoselect_result.returncode}:\n"
                + autoselect_result.stderr.decode("utf8", "replace")
            )
            rw.transport.nack(header)
            return

        # Find which classes were picked
        classes_block = cif.read_file(
            f"{select_dir}/{autoselect_params.classes_file}"
        ).sole_block()
        selected_classes = list(classes_block.find_loop("_rlnReferenceImage"))

        # Class ids get fed in as a string, need to convert these to a dictionary
        self.class_uuids_dict = json.loads(
            autoselect_params.class_uuids.replace("'", '"')
        )
        self.class_uuids_keys = list(self.class_uuids_dict.keys())

        # Send the picked classes to ispyb
        ispyb_command_list = []
        for picked_class in selected_classes:
            class_id = int(picked_class.split("@")[0])
            ispyb_command_list.append(
                {
                    "ispyb_command": "buffer",
                    "buffer_lookup": {
                        "particle_classification_id": self.class_uuids_dict[
                            self.class_uuids_keys[class_id - 1]
                        ],
                    },
                    "buffer_command": {
                        "ispyb_command": "insert_particle_classification"
                    },
                    "selected": 1,
                }
            )
        ispyb_parameters = {
            "ispyb_command": "multipart_message",
            "ispyb_command_list": ispyb_command_list,
        }
        self.log.info(f"Sending to ispyb {ispyb_parameters}")
        if isinstance(rw, MockRW):
            rw.transport.send(
                destination="ispyb_connector",
                message={
                    "parameters": ispyb_parameters,
                    "content": {"dummy": "dummy"},
                },
            )
        else:
            rw.send_to("ispyb_connector", ispyb_parameters)

        # Run the combine star files job to combine the files into particles_all.star
        self.log.info("Running star file combination and splitting")
        combine_star_dir = Path(
            project_dir / f"Select/job{autoselect_params.combine_star_job_number:03}"
        )
        files_to_combine = [select_dir / autoselect_params.particles_file]
        if (combine_star_dir / "particles_all.star").exists():
            files_to_combine.append(combine_star_dir / "particles_all.star")
        else:
            combine_star_dir.mkdir(parents=True, exist_ok=True)
            Path(project_dir / "Select/Best_particles").symlink_to(combine_star_dir)
            self.previous_total_count = 0

        if not (
            combine_star_dir / f".done_{autoselect_params.particles_file}"
        ).is_file():
            # Only run this if the particles file has not been added before
            combine_node_creator_params = {
                "job_type": "combine_star_files_job",
                "input_file": f"{select_dir}/{autoselect_params.particles_file}",
                "output_file": f"{combine_star_dir}/particles_all.star",
                "relion_options": dict(autoselect_params.relion_options),
                "command": (
                    "combine_star_files "
                    + " ".join([str(i) for i in files_to_combine])
                    + f" --output_dir {combine_star_dir}"
                ),
                "stdout": "",
                "stderr": "",
            }

            # Call the combining function and redirect prints to an io object
            combine_result = io.StringIO()
            with redirect_stdout(combine_result):
                try:
                    combine_star_files(files_to_combine, combine_star_dir)
                    (
                        combine_star_dir / f".done_{autoselect_params.particles_file}"
                    ).touch()
                    combine_node_creator_params["success"] = True
                except (IndexError, KeyError):
                    combine_node_creator_params["success"] = False
            self.parse_combiner_output(combine_result.getvalue())

            # Send combination job to node creator
            self.log.info("Sending combine_star_files_job (combine) to node creator")
            if isinstance(rw, MockRW):
                rw.transport.send(
                    destination="node_creator",
                    message={
                        "parameters": combine_node_creator_params,
                        "content": "dummy",
                    },
                )
            else:
                rw.send_to("node_creator", combine_node_creator_params)

            # End here if the command failed
            if not combine_node_creator_params["success"]:
                self.log.error("Star file combination failed")
                rw.transport.nack(header)
                return
        else:
            # If combination isn't run the number of particles needs to be found
            self.total_count = 0
            self.previous_total_count = 0
            with open(combine_star_dir / "particles_all.star", "r") as particles_file:
                while True:
                    particle_line = particles_file.readline()
                    if not particle_line:
                        break
                    particle_split_line = particle_line.split()
                    if (
                        len(particle_split_line) > 0
                        and particle_split_line[0][0].isdigit()
                    ):
                        self.total_count += 1
                        self.previous_total_count += 1

        # Create a file containing all selected classes
        if not (combine_star_dir / autoselect_params.classes_file).is_file():
            add_header = True
        else:
            add_header = False
        with open(select_dir / autoselect_params.classes_file, "r") as class_file, open(
            combine_star_dir / autoselect_params.classes_file, "a"
        ) as combined_classes:
            while True:
                line = class_file.readline()
                if not line:
                    break
                if line[0].isdigit():
                    # Always add lines which list a class
                    add_header = False
                    combined_classes.write(line)
                elif add_header:
                    # Only add other lines if this is a new file and not on classes yet
                    combined_classes.write(line)

        # Determine the next split size to use and whether to run 3D classification
        send_to_3d_classification = False
        if self.previous_total_count == 0:
            # First run of this job, use class3d_max_size
            next_batch_size = autoselect_params.class3d_batch_size
            if self.total_count > autoselect_params.class3d_batch_size:
                # Do 3D classification if there are more particles than the batch size
                send_to_3d_classification = True
        elif self.previous_total_count >= autoselect_params.class3d_max_size:
            # Iterations beyond those where 3D classification is run
            next_batch_size = autoselect_params.class3d_max_size
        else:
            # Re-runs with fewer particles than the maximum
            previous_batch_multiple = (
                self.previous_total_count // autoselect_params.class3d_batch_size
            )
            new_batch_multiple = (
                self.total_count // autoselect_params.class3d_batch_size
            )
            if new_batch_multiple > previous_batch_multiple:
                # Do 3D classification if a batch threshold has been crossed
                send_to_3d_classification = True
                # Set the batch size from the total count, but do not exceed the maximum
                next_batch_size = (
                    new_batch_multiple * autoselect_params.class3d_batch_size
                )
                if next_batch_size > autoselect_params.class3d_max_size:
                    next_batch_size = autoselect_params.class3d_max_size
            else:
                # Otherwise just get the next threshold
                next_batch_size = (
                    previous_batch_multiple + 1
                ) * autoselect_params.class3d_batch_size

        # Run the combine star files job to split particles_all.star into batches
        split_node_creator_params = {
            "job_type": "combine_star_files_job",
            "input_file": f"{select_dir}/{autoselect_params.particles_file}",
            "output_file": f"{combine_star_dir}/particles_all.star",
            "relion_options": dict(autoselect_params.relion_options),
            "command": (
                f"combine_star_files {combine_star_dir}/particles_all.star "
                f"--output_dir {combine_star_dir} "
                f"--split --split_size {next_batch_size}"
            ),
            "stdout": "",
            "stderr": "",
        }

        # Call the combining function and redirect prints to an io object
        split_result = io.StringIO()
        with redirect_stdout(split_result):
            try:
                split_star_file(
                    file_to_process=combine_star_dir / "particles_all.star",
                    output_dir=combine_star_dir,
                    split_size=next_batch_size,
                )
                split_node_creator_params["success"] = True
            except (IndexError, KeyError):
                split_node_creator_params["success"] = False
        self.parse_combiner_output(split_result.getvalue())

        # Send splitting job to node creator
        self.log.info("Sending combine_star_files_job (split) to node creator")
        if isinstance(rw, MockRW):
            rw.transport.send(
                destination="node_creator",
                message={"parameters": split_node_creator_params, "content": "dummy"},
            )
        else:
            rw.send_to("node_creator", split_node_creator_params)

        # End here if the command failed
        if not split_node_creator_params["success"]:
            self.log.error("Star file splitting failed")
            rw.transport.nack(header)
            return

        # Create 3D classification jobs
        if send_to_3d_classification:
            # Only send to 3D if a new multiple of the batch threshold is crossed
            # and the count has not passed the maximum
            self.log.info("Sending to Murfey for Class3D")

            # Copy the particle batch file
            shutil.copy2(
                combine_star_dir / "particles_split1.star",
                combine_star_dir / f"particles_batch_{next_batch_size}.star",
            )

            # Tell Murfey to do Class3D
            class3d_params = {
                "particles_file": f"{combine_star_dir}/particles_batch_{next_batch_size}.star",
                "class3d_dir": f"{project_dir}/Class3D/job",
                "batch_size": next_batch_size,
            }
            murfey_3d_params = {
                "register": "run_class3d",
                "class3d_message": class3d_params,
            }
            if isinstance(rw, MockRW):
                rw.transport.send("murfey_feedback", murfey_3d_params)
            else:
                rw.send_to("murfey_feedback", murfey_3d_params)

        murfey_confirmation = {
            "register": "done_class_selection",
        }
        if isinstance(rw, MockRW):
            rw.transport.send(
                destination="murfey_feedback", message=murfey_confirmation
            )
        else:
            rw.send_to("murfey_feedback", murfey_confirmation)

        # Remove the temporary hold file
        (combine_star_dir / f".done_{autoselect_params.particles_file}").unlink(
            missing_ok=True
        )

        self.log.info(f"Done {self.job_type} for {autoselect_params.input_file}.")
        rw.transport.ack(header)



from __future__ import annotations

import argparse
from math import ceil
from pathlib import Path
from typing import List

import starfile


def write_empty_particles_file(file_to_write, optics_dataframe, particles_dataframe):
    """Write a particles star file with no particles, ready for appending to"""
    with open(file_to_write, "w") as optics_file:
        optics_file.write("data_optics\n\nloop_\n")
        for optics_loop_tag in optics_dataframe.keys():
            optics_file.write(f"_{optics_loop_tag}\n")
        optics_file.write(" ".join(optics_dataframe.to_numpy(dtype=str)[0]))
        optics_file.write("\n\n\n")

        optics_file.write("data_particles\n\nloop_\n")
        for particles_loop_tag in particles_dataframe.keys():
            optics_file.write(f"_{particles_loop_tag}\n")


def combine_star_files(files_to_process: List[Path], output_dir: Path):
    """Combines any number of particle star files into a single file.

    Parameters:
    files_to_process: A list of the particle star files to combine
    output_dir: The directory in which to save the combined "particles_all.star" file
    """
    total_particles = 0

    # Never read particles_all.star first as it will be big
    if files_to_process[0].name == "particles_all.star":
        files_to_process.append(files_to_process[0])
        files_to_process = files_to_process[1:]

    # Make a temporary star file to get the table headings from
    reference_optics = None
    with open(files_to_process[0], "r") as full_starfile, open(
        output_dir / ".particles_tmp.star", "w"
    ) as tmp_starfile:
        for line_counter in range(50):
            line = full_starfile.readline()
            if line.startswith("opticsGroup"):
                reference_optics = line.split()
            if not line:
                break
            tmp_starfile.write(line)

    star_dictionary = starfile.read(output_dir / ".particles_tmp.star")
    (output_dir / ".particles_tmp.star").unlink()

    if not reference_optics:
        raise IndexError(f"Cannot find optics group in {files_to_process[0]}")

    write_empty_particles_file(
        output_dir / ".particles_all_tmp.star",
        star_dictionary["optics"],
        star_dictionary["particles"],
    )

    number_of_star_files = 0
    # Add the remaining files using append mode for speed and memory efficiency
    for split_file in files_to_process:
        # Check that the files have the same optics tables
        with open(split_file, "r") as added_starfile:
            while True:
                optics_line = added_starfile.readline()
                if not optics_line:
                    raise IndexError(f"Cannot find optics group in {split_file}")
                if optics_line.startswith("opticsGroup"):
                    new_optics = optics_line.split()
                    break

        if len(new_optics) != len(reference_optics):
            raise IndexError(
                "Cannot combine star files with different length optics tables."
            )
        for optics_label in range(len(reference_optics)):
            ref_value = reference_optics[optics_label]
            new_value = new_optics[optics_label]
            if ref_value[0].isdigit() and new_value[0].isdigit():
                ref_value = float(ref_value)
                new_value = float(new_value)
            if ref_value != new_value:
                print(ref_value, new_value)
                raise IndexError(
                    "Cannot combine star files with different values in optics tables."
                )

        # Add the particles lines to the final star file
        file_particles_count = 0
        with open(split_file, "r") as added_starfile, open(
            output_dir / ".particles_all_tmp.star", "a"
        ) as particles_file:
            while True:
                particle_line = added_starfile.readline()
                if not particle_line:
                    break
                particle_split_line = particle_line.split()
                if len(particle_split_line) > 0 and particle_split_line[0][0].isdigit():
                    file_particles_count += 1
                    total_particles += 1
                    particles_file.write(particle_line)

        print(f"Adding {split_file} with {file_particles_count} particles")
        number_of_star_files += 1

    (output_dir / ".particles_all_tmp.star").rename(output_dir / "particles_all.star")
    print(
        f"Combined {number_of_star_files} files into particles_all.star "
        f"with {total_particles} particles"
    )


def split_star_file(
    file_to_process: Path,
    output_dir: Path,
    number_of_splits: int or None = None,
    split_size: int or None = None,
):
    """Splits a star file into subfiles.

    The number of subfiles can be given with number_of_splits
    or is determined by split_size, the number of particles for in each file.
    """

    # Make a temporary star file to get the table headings from
    with open(file_to_process, "r") as full_starfile, open(
        output_dir / ".particles_tmp.star", "w"
    ) as tmp_starfile:
        for line_counter in range(50):
            line = full_starfile.readline()
            if not line:
                break
            tmp_starfile.write(line)

    star_dictionary = starfile.read(output_dir / ".particles_tmp.star")
    (output_dir / ".particles_tmp.star").unlink()

    # Find the number of lines in the full file
    starfile_starter_lines = line_counter - star_dictionary["particles"].shape[0]
    count = 0
    with open(file_to_process, "r") as full_starfile:
        for count, line in enumerate(full_starfile):
            pass
    number_of_particles = count + 1 - starfile_starter_lines

    # Determine the number of files and size of the splits
    if number_of_splits:
        if split_size:
            print(
                "Warning: "
                "Both number_of_splits and split_size have been given, "
                "using number_of_splits.",
            )
        split_size = ceil(number_of_particles / number_of_splits)
    elif split_size:
        number_of_splits = ceil(number_of_particles / split_size)
    else:
        raise KeyError("Either number_of_splits or split_size must be given.")

    with open(file_to_process, "r") as full_starfile:
        # Read in the full file line by line, removing the header lines first
        for start_line in range(starfile_starter_lines):
            full_starfile.readline()

        for split in range(number_of_splits):
            # Give each new file the header information
            write_empty_particles_file(
                output_dir / f".particles_split{split+1}_tmp.star",
                star_dictionary["optics"],
                star_dictionary["particles"],
            )

            # Write particles to the split files by reading in lines from the full file
            with open(
                output_dir / f".particles_split{split+1}_tmp.star", "a"
            ) as split_file:
                for count in range(split_size):
                    particle_line = full_starfile.readline()
                    if not particle_line:
                        break
                    split_file.write(particle_line)

            (output_dir / f".particles_split{split+1}_tmp.star").rename(
                output_dir / f"particles_split{split+1}.star"
            )

    print(
        f"Split {number_of_particles} particles into "
        f"{number_of_splits} files with {split_size} particles in each."
    )


def create_parser():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "files_to_process",
        type=Path,
        nargs="+",
        help="The star files from which to combine particles.",
    )
    parser.add_argument(
        "--output_dir",
        dest="output_dir",
        type=Path,
        help="Folder in which to save the new star files.",
    )

    parser.add_argument(
        "--split",
        dest="do_split",
        action="store_true",
        default=False,
        help="Whether to split the particles again into new star files.",
    )
    parser.add_argument(
        "--n_files",
        dest="n_files",
        type=int,
        default=None,
        help="Number of files to split the particles into.",
    )
    parser.add_argument(
        "--split_size",
        dest="split_size",
        type=int,
        default=None,
        help="Number of particles to write per file.",
    )

    return parser


def main():
    arg_parser = create_parser()
    run_args = vars(arg_parser.parse_args())

    combine_star_files(run_args["files_to_process"], run_args["output_dir"])

    if run_args["do_split"]:
        split_star_file(
            run_args["output_dir"] / "particles_all.star",
            run_args["output_dir"],
            run_args["n_files"],
            run_args["split_size"],
        )



from __future__ import annotations

import json
import os
import string
import subprocess
import time
from collections import ChainMap
from pathlib import Path

import yaml
from workflows.services.common_service import CommonService

from cryoemservices.services.motioncorr import MotionCorr

"""
This service submits motion correction jobs to a slurm cluster
To do this it needs environment variables set for the following:
    MOTIONCOR2_SIF: singularity image of MotionCor2
    SLURM_RESTAPI_CONFIG: configuration yaml file for the slurm cluster

The configuration has the following format:
    plugin: slurm
    url: <url>:<port>
    user_token: <file with restapi token>
    user: <username>
    user_home: <home directory>
    api_version: v0.0.38 or v0.0.40
    partition: <optional slurm partition>
    partition_preference: <optional slurm preferences>
    clusters: <optional slurm clusters>
    required_directories: [<list of directories to bind for singularity>]
"""

slurm_json_job_template = {
    "v0.0.38": {
        "name": "MotionCorr",
        "nodes": 1,
        "tasks": 1,
        "cpus_per_task": 1,
        "gpus": 1,
        "memory_per_gpu": 12000,
        "time_limit": "1:00:00",
    },
    "v0.0.40": {
        "name": "MotionCorr",
        "minimum_nodes": 1,
        "maximum_nodes": 1,
        "tasks": 1,
        "cpus_per_task": 1,
        "tres_per_task": "gres/gpu:1",
        "memory_per_node": {
            "number": 12000,
            "set": True,
            "infinite": False,
        },
        "time_limit": {
            "number": 3600,
            "set": True,
            "infinite": False,
        },
    },
}
slurm_script_template = (
    "#!/bin/bash\n"
    "echo \"$(date '+%Y-%m-%d %H:%M:%S.%3N'): running MotionCor2\"\n"
    "mkdir /tmp/tmp_$SLURM_JOB_ID\n"
    "export APPTAINER_CACHEDIR=/tmp/tmp_$SLURM_JOB_ID\n"
    "export APPTAINER_TMPDIR=/tmp/tmp_$SLURM_JOB_ID\n"
    "singularity exec --nv --bind /lib64,/tmp/tmp_$SLURM_JOB_ID:/tmp"
)
slurm_tmp_cleanup = "\nrm -rf /tmp/tmp_$SLURM_JOB_ID"


class ChainMapWithReplacement(ChainMap):
    def __init__(self, *maps, substitutions=None) -> None:
        super().__init__(*maps)
        self._substitutions = substitutions

    def __getitem__(self, k):
        v = super().__getitem__(k)
        if self._substitutions and isinstance(v, str) and "$" in v:
            template = string.Template(v)
            return template.substitute(**self._substitutions)
        return v


class MotionCorrSlurm(MotionCorr, CommonService):
    """
    A service for submitting MotionCor2 jobs to slurm via a RestAPI
    """

    # Logger name
    _logger_name = "cryoemservices.services.motioncorr_slurm"

    def parse_mc_slurm_output(self, mc_output_file):
        """
        Read the output logs of MotionCorr to determine
        the movement of each frame
        """
        with open(mc_output_file, "r") as mc_file:
            lines = mc_file.readlines()
            frames_line = False
            for line in lines:
                # Frame reading in MotionCorr 1.4.0
                if line.startswith("...... Frame"):
                    line_split = line.split()
                    self.x_shift_list.append(float(line_split[-2]))
                    self.y_shift_list.append(float(line_split[-1]))

                # Alternative frame reading for MotionCorr 1.6.3
                if not line:
                    frames_line = False
                if frames_line:
                    line_split = line.split()
                    self.x_shift_list.append(float(line_split[1]))
                    self.y_shift_list.append(float(line_split[2]))
                if "x Shift" in line:
                    frames_line = True

    def motioncor2(self, command: list, mrc_out: Path):
        """Submit MotionCor2 jobs to the slurm cluster via the RestAPI"""
        try:
            # Get the configuration and token for the restAPI
            with open(os.environ["SLURM_RESTAPI_CONFIG"], "r") as f:
                slurm_rest = yaml.safe_load(f)
            user = slurm_rest["user"]
            user_home = slurm_rest["user_home"]
            with open(slurm_rest["user_token"], "r") as f:
                slurm_token = f.read().strip()
        except (KeyError, FileNotFoundError):
            self.log.error("Unable to load slurm restAPI config file and token")
            return subprocess.CompletedProcess(
                args="",
                returncode=1,
                stdout="".encode("utf8"),
                stderr="No restAPI config or token".encode("utf8"),
            )

        # Check the API version is one this service has been tested with
        api_version = slurm_rest["api_version"]
        print(api_version)
        if api_version not in ["v0.0.38", "v0.0.40"]:
            return subprocess.CompletedProcess(
                args="",
                returncode=1,
                stdout="".encode("utf8"),
                stderr=f"Unsupported API version {api_version}".encode("utf8"),
            )

        # Construct the json for submission
        mc_output_file = f"{mrc_out}.out"
        mc_error_file = f"{mrc_out}.err"
        submission_file = f"{mrc_out}.json"
        slurm_config = {
            "standard_output": mc_output_file,
            "standard_error": mc_error_file,
            "current_working_directory": str(Path(mrc_out).parent),
        }
        if api_version == "v0.0.38":
            slurm_config["environment"] = {"USER": user, "HOME": user_home}
        else:
            slurm_config["environment"] = [f"USER: {user}", f"HOME: {user_home}"]

        # Add slurm partition and cluster preferences if given
        if slurm_rest.get("partition"):
            slurm_config["partition"] = slurm_rest["partition"]
        if slurm_rest.get("partition_preference"):
            slurm_config["prefer"] = slurm_rest["partition_preference"]
        if slurm_rest.get("clusters"):
            slurm_config["clusters"] = slurm_rest["clusters"]
        # Combine this with the template for the given API version
        slurm_json_job = dict(slurm_json_job_template[api_version], **slurm_config)

        # Make the script command and save the submission json
        if slurm_rest.get("required_directories"):
            binding_dirs = "," + ",".join(slurm_rest["required_directories"])
        else:
            binding_dirs = ""
        job_command = (
            slurm_script_template
            + f"{binding_dirs} --home {user_home} {os.environ['MOTIONCOR2_SIF']} "
            + " ".join(command)
            + slurm_tmp_cleanup
        )
        slurm_json = {"job": slurm_json_job, "script": job_command}
        with open(submission_file, "w") as f:
            json.dump(slurm_json, f)

        # Command to submit jobs to the restAPI
        slurm_submit_command = (
            f'curl -H "X-SLURM-USER-NAME:{user}" -H "X-SLURM-USER-TOKEN:{slurm_token}" '
            '-H "Content-Type: application/json" -X POST '
            f'{slurm_rest["url"]}/slurm/{slurm_rest["api_version"]}/job/submit '
            f"-d @{submission_file}"
        )
        slurm_submission_json = subprocess.run(
            slurm_submit_command, capture_output=True, shell=True
        )
        try:
            # Extract the job id from the submission response to use in the next query
            slurm_response = slurm_submission_json.stdout.decode("utf8", "replace")
            slurm_response_json = json.loads(slurm_response)
            job_id = slurm_response_json["job_id"]
        except (json.JSONDecodeError, KeyError):
            self.log.error(
                f"Unable to submit job to {slurm_rest['url']}. The restAPI returned "
                f"{slurm_submission_json.stdout.decode('utf8', 'replace')}"
            )
            return subprocess.CompletedProcess(
                args="",
                returncode=1,
                stdout=slurm_submission_json.stdout,
                stderr=slurm_submission_json.stderr,
            )
        self.log.info(f"Submitted MotionCorr job {job_id} to slurm. Waiting...")
        if slurm_response_json.get("warnings") and slurm_response_json["warnings"]:
            self.log.warning(
                f"Slurm reported these warnings: {slurm_response_json['warnings']}"
            )
        if slurm_response_json.get("errors") and slurm_response_json["errors"]:
            self.log.warning(
                f"Slurm reported these errors: {slurm_response_json['errors']}"
            )

        # Command to get the status of the submitted job from the restAPI
        slurm_status_command = (
            f'curl -H "X-SLURM-USER-NAME:{user}" -H "X-SLURM-USER-TOKEN:{slurm_token}" '
            '-H "Content-Type: application/json" -X GET '
            f'{slurm_rest["url"]}/slurm/{slurm_rest["api_version"]}/job/{job_id}'
        )
        slurm_job_state = "PENDING"

        # Wait until the job has a status indicating it has finished
        loop_counter = 0
        while slurm_job_state in (
            "PENDING",
            "CONFIGURING",
            "RUNNING",
            "COMPLETING",
        ):
            if loop_counter < 5:
                time.sleep(5)
            else:
                time.sleep(30)
            loop_counter += 1

            # Call the restAPI to find out the job state
            slurm_status_json = subprocess.run(
                slurm_status_command, capture_output=True, shell=True
            )
            try:
                slurm_response = slurm_status_json.stdout.decode("utf8", "replace")
                slurm_job_state = json.loads(slurm_response)["jobs"][0]["job_state"]
                if api_version == "v0.0.40":
                    slurm_job_state = slurm_job_state[0]
            except (json.JSONDecodeError, IndexError, KeyError):
                print(slurm_status_command)
                self.log.error(
                    f"Unable to get status for job {job_id}. The restAPI returned "
                    f"{slurm_status_json.stdout.decode('utf8', 'replace')}"
                )
                return subprocess.CompletedProcess(
                    args="",
                    returncode=1,
                    stdout=slurm_status_json.stdout,
                    stderr=slurm_status_json.stderr,
                )

            if loop_counter >= 60:
                slurm_cancel_command = (
                    f'curl -H "X-SLURM-USER-NAME:{user}" '
                    f'-H "X-SLURM-USER-TOKEN:{slurm_token}" '
                    '-H "Content-Type: application/json" -X DELETE '
                    f'{slurm_rest["url"]}/slurm/{slurm_rest["api_version"]}/job/{job_id}'
                )
                subprocess.run(slurm_cancel_command, capture_output=True, shell=True)
                self.log.error("Timeout running motion correction")
                return subprocess.CompletedProcess(
                    args="",
                    returncode=1,
                    stdout="".encode("utf8"),
                    stderr="Timeout running motion correction".encode("utf8"),
                )

        # Read in the MotionCor output then clean up the files
        self.log.info(f"Job {job_id} has finished!")
        try:
            self.parse_mc_slurm_output(mc_output_file)
            with open(mc_output_file, "r") as mc_stdout:
                stdout = mc_stdout.read()
            with open(mc_error_file, "r") as mc_stderr:
                stderr = mc_stderr.read()
        except FileNotFoundError:
            self.log.error(f"MotionCor output file {mc_output_file} not found")
            stdout = ""
            stderr = f"Reading MotionCor output file {mc_output_file} failed"
            slurm_job_state = "FAILED"

        if self.x_shift_list and self.y_shift_list:
            Path(mc_output_file).unlink()
            Path(mc_error_file).unlink()
            Path(submission_file).unlink()
        else:
            self.log.error(f"Reading shifts from {mc_output_file} failed")
            slurm_job_state = "FAILED"

        if slurm_job_state == "COMPLETED":
            return subprocess.CompletedProcess(
                args="",
                returncode=0,
                stdout=stdout.encode("utf8"),
                stderr=stderr.encode("utf8"),
            )
        else:
            self.x_shift_list = []
            self.y_shift_list = []
            return subprocess.CompletedProcess(
                args="",
                returncode=1,
                stdout=stdout.encode("utf8"),
                stderr=stderr.encode("utf8"),
            )



from __future__ import annotations

import string
import time
from collections import ChainMap
from datetime import datetime
from pathlib import Path

import ispyb
import ispyb.sqlalchemy as models
import sqlalchemy.exc
import sqlalchemy.orm
import workflows.recipe
from pydantic import BaseModel, validate_arguments
from workflows.services.common_service import CommonService

import cryoemservices.services.ispyb_buffer as buffer


class ChainMapWithReplacement(ChainMap):
    def __init__(self, *maps, substitutions=None) -> None:
        super().__init__(*maps)
        self._substitutions = substitutions

    def __getitem__(self, k):
        v = super().__getitem__(k)
        if self._substitutions and isinstance(v, str) and "$" in v:
            template = string.Template(v)
            return template.substitute(**self._substitutions)
        return v


class MovieParams(BaseModel):
    dcid: int
    movie_number: int = None  # image number
    movie_path: str = None  # micrograph full path
    timestamp: float = None


def lookup_command(command, refclass):
    return getattr(refclass, "do_" + command, None)


class EMISPyB(CommonService):
    """A service that receives information to be written to ISPyB."""

    # Human readable service name
    _service_name = "EMISPyB"

    # Logger name
    _logger_name = "cryoemservices.services.ispyb"

    # ispyb connection details
    ispyb = None
    _ispyb_sessionmaker = None

    def initializing(self):
        """Subscribe the ISPyB connector queue. Received messages must be
        acknowledged. Prepare ISPyB database connection."""
        self.log.info(f"ISPyB connector using ispyb v{ispyb.__version__}")
        self.ispyb = ispyb.open()
        self._ispyb_sessionmaker = sqlalchemy.orm.sessionmaker(
            bind=sqlalchemy.create_engine(
                ispyb.sqlalchemy.url(), connect_args={"use_pure": True}
            )
        )
        try:
            self.log.info("Cleaning up ISPyB buffer table...")
            with self._ispyb_sessionmaker() as session:
                buffer.evict(session=session)
        except Exception as e:
            self.log.warning(
                f"Encountered exception {e!r} while cleaning up ISPyB buffer table",
                exc_info=True,
            )
        self.log.info("ISPyB service ready")
        workflows.recipe.wrap_subscribe(
            self._transport,
            "ispyb_connector",
            self.receive_msg,
            acknowledgement=True,
            log_extender=self.extend_log,
            allow_non_recipe_messages=True,
        )

    def receive_msg(self, rw, header, message):
        """Do something with ISPyB."""

        if header.get("redelivered") == "true":
            # A redelivered message may just have been processed in a parallel instance,
            # which was connected to a different database server in the DB cluster. If
            # we were to process it immediately we may run into a DB synchronization
            # fault. Avoid this by giving the DB cluster a bit of time to settle.
            self.log.debug("Received redelivered message, holding for a moment.")
            time.sleep(0.5)

        if not rw:
            # Incoming message is not a recipe message. Simple messages can be valid
            if (
                not isinstance(message, dict)
                or not message.get("parameters")
                or not message.get("content")
            ):
                self.log.error("Rejected invalid simple message")
                self._transport.nack(header)
                return
            self.log.debug("Received a simple message")

            # Create a wrapper-like object that can be passed to functions
            # as if a recipe wrapper was present.
            class MockRW:
                def dummy(self, *args, **kwargs):
                    pass

            rw = MockRW()
            rw.transport = self._transport
            rw.recipe_step = {"parameters": message["parameters"]}
            rw.environment = {"has_recipe_wrapper": False}
            rw.set_default_channel = rw.dummy
            rw.send = rw.dummy
            message = message["content"]

        command = rw.recipe_step["parameters"].get("ispyb_command")
        if not command:
            self.log.error("Received message is not a valid ISPyB command")
            rw.transport.nack(header)
            return
        command_function = lookup_command(command, self)
        if not command_function:
            self.log.error("Received unknown ISPyB command (%s)", command)
            rw.transport.nack(header)
            return

        self.log.debug("Running ISPyB call %s", command)
        txn = rw.transport.transaction_begin(subscription_id=header["subscription"])
        rw.set_default_channel("output")

        parameter_map = ChainMapWithReplacement(
            message if isinstance(message, dict) else {},
            rw.recipe_step["parameters"],
            substitutions=rw.environment,
        )

        def parameters(parameter, replace_variables=True):
            if isinstance(message, dict):
                base_value = message.get(
                    parameter, rw.recipe_step["parameters"].get(parameter)
                )
            else:
                base_value = rw.recipe_step["parameters"].get(parameter)
            if (
                not replace_variables
                or not base_value
                or not isinstance(base_value, str)
                or "$" not in base_value
            ):
                return base_value
            for key in sorted(rw.environment, key=len, reverse=True):
                if "${" + key + "}" in base_value:
                    base_value = base_value.replace(
                        "${" + key + "}", str(rw.environment[key])
                    )
                # Replace longest keys first, as the following replacement is
                # not well-defined when one key is a prefix of another:
                if "$" + key in base_value:
                    base_value = base_value.replace("$" + key, str(rw.environment[key]))
            return base_value

        try:
            with self._ispyb_sessionmaker() as session:
                result = command_function(
                    rw=rw,
                    message=message,
                    parameters=parameters,
                    parameter_map=parameter_map,
                    session=session,
                    transaction=txn,
                    header=header,
                )
        except Exception as e:
            self.log.error(
                f"Uncaught exception {e!r} in ISPyB function {command!r}, "
                "quarantining message and shutting down instance.",
                exc_info=True,
            )
            rw.transport.transaction_abort(txn)
            rw.transport.nack(header)
            self._request_termination()
            return

        store_result = rw.recipe_step["parameters"].get("store_result")
        if store_result and result and "return_value" in result:
            rw.environment[store_result] = result["return_value"]
            self.log.debug(
                "Storing result '%s' in environment variable '%s'",
                result["return_value"],
                store_result,
            )
        if result and result.get("success"):
            rw.send({"result": result.get("return_value")}, transaction=txn)
            rw.transport.ack(header, transaction=txn)
        elif result and result.get("checkpoint") and not result.get("delay"):
            rw.checkpoint(
                result.get("return_value"),
                delay=rw.recipe_step["parameters"].get("delay", result.get("delay")),
                transaction=txn,
            )
            rw.transport.ack(header, transaction=txn)
        else:
            rw.transport.transaction_abort(txn)
            rw.transport.nack(header)
            return
        rw.transport.transaction_commit(txn)

    def do_multipart_message(self, rw, message, **kwargs):
        """The multipart_message command allows the recipe or client to specify a
        multi-stage operation. With this you can process a list of API calls.
        Each API call may have a return value that can be stored.
        Multipart_message takes care of chaining and checkpointing to make the
        overall call near-ACID compliant."""

        if not rw.environment.get("has_recipe_wrapper", True):
            self.log.error(
                "Multipart message call can not be used with simple messages"
            )
            return False

        step = 1
        commands = rw.recipe_step["parameters"].get("ispyb_command_list")
        if isinstance(message, dict) and isinstance(
            message.get("ispyb_command_list"), list
        ):
            commands = message["ispyb_command_list"]
            step = message.get("checkpoint", 0) + 1
        if not commands:
            self.log.error("Received multipart message containing no commands")
            return False

        current_command = commands[0]
        command = current_command.get("ispyb_command")
        if not command:
            self.log.error(
                "Multipart command %s is not a valid ISPyB command", current_command
            )
            return False
        command_function = lookup_command(command, self)
        if not command_function:
            self.log.error("Received unknown ISPyB command (%s)", command)
            return False
        self.log.debug(
            "Processing step %d of multipart message (%s) with %d further steps",
            step,
            command,
            len(commands) - 1,
            extra={"ispyb-message-parts": len(commands)} if step == 1 else {},
        )

        # Create a parameter lookup function specific to this step of the
        # multipart message
        def step_parameters(parameter, replace_variables=True):
            """Slight change in behaviour compared to 'parameters' in a direct call:
            If the value is defined in the command list item then this takes
            precedence. Otherwise we check the original message content. Finally,
            we look in the parameters dictionary of the recipe step for the
            multipart_message command.
            String replacement rules apply as usual."""
            if parameter in current_command:
                base_value = current_command[parameter]
            elif isinstance(message, dict) and parameter in message:
                base_value = message[parameter]
            else:
                base_value = rw.recipe_step["parameters"].get(parameter)
            if (
                not replace_variables
                or not base_value
                or not isinstance(base_value, str)
                or "$" not in base_value
            ):
                return base_value
            for key in sorted(rw.environment, key=len, reverse=True):
                if "${" + key + "}" in base_value:
                    base_value = base_value.replace(
                        "${" + key + "}", str(rw.environment[key])
                    )
                # Replace longest keys first, as the following replacement is
                # not well-defined when one key is a prefix of another:
                if "$" + key in base_value:
                    base_value = base_value.replace("$" + key, str(rw.environment[key]))
            return base_value

        kwargs["parameters"] = step_parameters

        # If this step previously checkpointed then override the message passed
        # to the step.
        step_message = current_command
        if isinstance(message, dict):
            step_message = message.get("step_message", step_message)

        # Run the multipart step
        result = command_function(rw=rw, message=step_message, **kwargs)

        # Store step result if appropriate
        store_result = current_command.get("store_result")
        if store_result and result and "return_value" in result:
            rw.environment[store_result] = result["return_value"]
            self.log.debug(
                "Storing result '%s' in environment variable '%s'",
                result["return_value"],
                store_result,
            )

        # If the current step has checkpointed then need to manage this
        if result and result.get("checkpoint"):
            self.log.debug("Checkpointing for sub-command %s", command)

            if isinstance(message, dict):
                checkpoint_dictionary = message
            else:
                checkpoint_dictionary = {}
            checkpoint_dictionary["checkpoint"] = step - 1
            checkpoint_dictionary["ispyb_command_list"] = commands
            checkpoint_dictionary["step_message"] = result.get("return_value")
            return {
                "checkpoint": True,
                "return_value": checkpoint_dictionary,
                "delay": result.get("delay"),
            }

        # If the step did not succeed then propagate failure
        if not result or not result.get("success"):
            self.log.debug("Multipart command failed")
            return result

        # Step has completed, so remove from queue
        commands.pop(0)

        # If the multipart command is finished then propagate success
        if not commands:
            self.log.debug("and done.")
            return result

        # If there are more steps then checkpoint the current state
        # and put it back on the queue (with no delay)
        self.log.debug("Checkpointing remaining %d steps", len(commands))
        if isinstance(message, dict):
            checkpoint_dictionary = message
        else:
            checkpoint_dictionary = {}
        checkpoint_dictionary["checkpoint"] = step
        checkpoint_dictionary["ispyb_command_list"] = commands
        if "step_message" in checkpoint_dictionary:
            del checkpoint_dictionary["step_message"]
        return {"checkpoint": True, "return_value": checkpoint_dictionary}

    def do_buffer(self, rw, message, session, parameters, header, **kwargs):
        """The buffer command supports running buffer lookups before running
        a command, and optionally storing the result in a buffer after running
        the command. It also takes care of checkpointing in case a required
        buffer value is not yet available.

        As an example, if you want to send this message to the ISPyB service:

        {
            "ispyb_command": "insert_thing",
            "parent_id": "$ispyb_thing_parent_id",
            "store_result": "ispyb_thing_id",
            "parameter_a": ...,
            "parameter_b": ...,
        }

        and want to look up the parent_id using the buffer with your unique
        reference UUID1 you could write:

        {
            "ispyb_command": "buffer",
            "program_id": "$ispyb_autoprocprogram_id",
            "buffer_lookup": {
                "parent_id": UUID1,
            },
            "buffer_command": {
                "ispyb_command": "insert_thing",
                "parameter_a": ...,
                "parameter_b": ...,
            },
            "buffer_store": UUID2,
            "store_result": "ispyb_thing_id",
        }

        which would also store the result under buffer reference UUID2.
        """

        if not rw.environment.get("has_recipe_wrapper", True):
            self.log.error("Buffer call can not be used with simple messages")
            return False

        if not isinstance(message, dict):
            self.log.error("Invalid buffer call: message must be a dictionary")
            return False

        if not isinstance(message.get("buffer_command"), dict) or not message[
            "buffer_command"
        ].get("ispyb_command"):
            self.log.error("Invalid buffer call: no buffer command specified")
            return False

        command_function = lookup_command(
            message["buffer_command"]["ispyb_command"], self
        )
        if not command_function:
            self.log.error("Invalid buffer call: unknown command specified")
            return False

        if ("buffer_expiry_time" not in message) or (
            header.get("dlq-reinjected") in {True, "True", "true", 1}
        ):
            message["buffer_expiry_time"] = time.time() + 600

        # Prepare command: Resolve all references
        program_id = parameters("program_id")
        if message.get("buffer_lookup"):
            if not isinstance(message["buffer_lookup"], dict):
                self.log.error(
                    "Invalid buffer call: buffer_lookup dictionary is not a dictionary"
                )
                return False
            if not program_id:
                self.log.error("Invalid buffer call: program_id is undefined")
                return False
            for entry in list(message["buffer_lookup"]):
                buffer_result = buffer.load(
                    session=session,
                    program=program_id,
                    uuid=message["buffer_lookup"][entry],
                )
                if buffer_result.success:
                    # resolve value and continue
                    message["buffer_command"][entry] = buffer_result.value
                    del message["buffer_lookup"][entry]
                    self.log.debug(
                        f"Successfully resolved buffer reference {entry!r} to {buffer_result.value!r}"
                    )
                    continue

                if message["buffer_expiry_time"] < time.time():
                    self.log.warning(
                        f"Buffer call could not be resolved: entry {entry} not found for program {program_id}"
                    )
                    return False

                # value can not yet be resolved, put request back in the queue
                return {"checkpoint": True, "return_value": message, "delay": 20}

        # Run the actual command
        result = command_function(
            rw=rw,
            message=message["buffer_command"],
            session=session,
            parameters=parameters,
            **kwargs,
        )

        # Store result if appropriate
        store_result = message.get("store_result")
        if store_result and result and "return_value" in result:
            rw.environment[store_result] = result["return_value"]
            self.log.debug(
                "Storing result '%s' in environment variable '%s'",
                result["return_value"],
                store_result,
            )

        # If the actual command has checkpointed then need to manage this
        if result and result.get("checkpoint"):
            self.log.debug("Checkpointing for buffered function")
            message["buffer_command"] = result["return_value"]
            return {
                "checkpoint": True,
                "return_value": message,
                "delay": result.get("delay"),
            }

        # If the command did not succeed then propagate failure
        if not result or not result.get("success"):
            self.log.warning("Buffered command failed")
            # to become debug level eventually, the actual function will do the warning
            return result

        # Optionally store a reference to the result in the buffer table
        if message.get("buffer_store"):
            self.log.debug("Storing buffer result for UUID %r", message["buffer_store"])
            buffer.store(
                session=session,
                program=program_id,
                uuid=message["buffer_store"],
                reference=result["return_value"],
            )

        # Finally, propagate result
        return result

    # EM-specific parts from here
    def _get_movie_id(
        self,
        full_path,
        data_collection_id,
        db_session,
    ):
        self.log.info(
            f"Looking for Movie ID. Movie name: {full_path} DCID: {data_collection_id}"
        )
        movie_name = str(Path(full_path).stem).replace("_motion_corrected", "")
        mv_query = db_session.query(models.Movie).filter(
            models.Movie.dataCollectionId == data_collection_id,
        )
        results = mv_query.all()
        correct_result = None
        if results:
            for result in results:
                if movie_name in result.movieFullPath:
                    correct_result = result
        if correct_result:
            mvid = correct_result.movieId
            self.log.info(f"Found Movie ID: {mvid}")
            return mvid
        else:
            self.log.error(f"Unable to find movie ID for {movie_name}")
            return None

    @validate_arguments(config={"arbitrary_types_allowed": True})
    def do_insert_movie(self, parameter_map: MovieParams, session, **kwargs):
        self.log.info("Inserting Movie parameters.")

        try:
            if parameter_map.timestamp:
                values = models.Movie(
                    dataCollectionId=parameter_map.dcid,
                    movieNumber=parameter_map.movie_number,
                    movieFullPath=parameter_map.movie_path,
                    createdTimeStamp=datetime.fromtimestamp(
                        parameter_map.timestamp
                    ).strftime("%Y-%m-%d %H:%M:%S"),
                )
            else:
                values = models.Movie(
                    dataCollectionId=parameter_map.dcid,
                    movieNumber=parameter_map.movie_number,
                    movieFullPath=parameter_map.movie_path,
                )
            session.add(values)
            session.commit()
            self.log.info(f"Created Movie record {values.movieId}")
            return {"success": True, "return_value": values.movieId}
        except sqlalchemy.exc.SQLAlchemyError as e:
            self.log.error(
                "Inserting movie entry caused exception '%s'.",
                e,
                exc_info=True,
            )
            return False

    def do_insert_motion_correction(self, parameters, session, message=None, **kwargs):
        if message is None:
            message = {}
        self.log.info("Inserting Motion Correction parameters.")

        def full_parameters(param):
            return message.get(param) or parameters(param)

        try:
            movie_id = None
            if full_parameters("movie_id") is None:
                movie_values = self.do_insert_movie(
                    parameter_map=MovieParams(
                        dcid=full_parameters("dcid"),
                        movie_number=full_parameters("image_number"),
                        movie_path=full_parameters("micrograph_full_path"),
                        timestamp=full_parameters("created_time_stamp"),
                    ),
                    session=session,
                )
                movie_id = movie_values["return_value"]

            values = models.MotionCorrection(
                dataCollectionId=full_parameters("dcid"),
                movieId=full_parameters("movie_id") or movie_id,
                autoProcProgramId=full_parameters("program_id"),
                imageNumber=full_parameters("image_number"),
                firstFrame=full_parameters("first_frame"),
                lastFrame=full_parameters("last_frame"),
                dosePerFrame=full_parameters("dose_per_frame"),
                doseWeight=full_parameters("dose_weight"),
                totalMotion=full_parameters("total_motion"),
                averageMotionPerFrame=full_parameters("average_motion_per_frame"),
                driftPlotFullPath=full_parameters("drift_plot_full_path"),
                micrographFullPath=full_parameters("micrograph_full_path"),
                micrographSnapshotFullPath=full_parameters(
                    "micrograph_snapshot_full_path"
                ),
                patchesUsedX=full_parameters("patches_used_x"),
                patchesUsedY=full_parameters("patches_used_y"),
                fftFullPath=full_parameters("fft_full_path"),
                fftCorrectedFullPath=full_parameters("fft_corrected_full_path"),
                comments=full_parameters("comments"),
            )
            session.add(values)
            session.commit()
            self.log.info(
                f"Created MotionCorrection record {values.motionCorrectionId}"
            )
            return {"success": True, "return_value": values.motionCorrectionId}
        except sqlalchemy.exc.SQLAlchemyError as e:
            self.log.error(
                "Inserting motion correction entry caused exception '%s'.",
                e,
                exc_info=True,
            )
            return False

    def do_insert_relative_ice_thickness(
        self, parameters, session, message=None, **kwargs
    ):
        if message is None:
            message = {}
        dcid = parameters("dcid")
        self.log.info(f"Inserting Relative Ice Thickness parameters. DCID: {dcid}")

        def full_parameters(param):
            return message.get(param) or parameters(param)

        try:
            values = models.RelativeIceThickness(
                motionCorrectionId=full_parameters("motion_correction_id"),
                autoProcProgramId=full_parameters("program_id"),
                minimum=full_parameters("minimum"),
                q1=full_parameters("q1"),
                median=full_parameters("median"),
                q3=full_parameters("q3"),
                maximum=full_parameters("maximum"),
            )
            session.add(values)
            session.commit()
            return {"success": True, "return_value": values.relativeIceThicknessId}
        except sqlalchemy.exc.SQLAlchemyError as e:
            self.log.error(
                "Inserting relative ice thickness entry caused exception '%s'.",
                e,
                exc_info=True,
            )
            return False

    def do_insert_ctf(self, parameters, session, message=None, **kwargs):
        if message is None:
            message = {}
        dcid = parameters("dcid")
        self.log.info(f"Inserting CTF parameters. DCID: {dcid}")

        def full_parameters(param):
            return message.get(param) or parameters(param)

        try:
            values = models.CTF(
                ctfId=full_parameters("ctf_id"),
                motionCorrectionId=full_parameters("motion_correction_id"),
                autoProcProgramId=full_parameters("program_id"),
                boxSizeX=full_parameters("box_size_x"),
                boxSizeY=full_parameters("box_size_y"),
                minResolution=full_parameters("min_resolution"),
                maxResolution=full_parameters("max_resolution"),
                minDefocus=full_parameters("min_defocus"),
                maxDefocus=full_parameters("max_defocus"),
                defocusStepSize=full_parameters("defocus_step_size"),
                astigmatism=full_parameters("astigmatism"),
                astigmatismAngle=full_parameters("astigmatism_angle"),
                estimatedResolution=full_parameters("estimated_resolution"),
                estimatedDefocus=full_parameters("estimated_defocus"),
                amplitudeContrast=full_parameters("amplitude_contrast"),
                ccValue=full_parameters("cc_value"),
                fftTheoreticalFullPath=full_parameters("fft_theoretical_full_path"),
                comments=full_parameters("comments"),
            )
            session.add(values)
            session.commit()
            self.log.info(f"Created CTF record {values.ctfId} for DCID {dcid}")
            return {"success": True, "return_value": values.ctfId}
        except sqlalchemy.exc.SQLAlchemyError as e:
            self.log.error(
                "Inserting CTF entry caused exception '%s'.",
                e,
                exc_info=True,
            )
            return False

    def do_insert_particle_picker(self, parameters, session, message=None, **kwargs):
        if message is None:
            message = {}
        dcid = parameters("dcid")
        self.log.info(f"Inserting Particle Picker parameters. DCID: {dcid}")

        def full_parameters(param):
            return message.get(param) or parameters(param)

        try:
            values = models.ParticlePicker(
                particlePickerId=full_parameters("particle_picker_id"),
                programId=full_parameters("program_id"),
                firstMotionCorrectionId=full_parameters("motion_correction_id"),
                particlePickingTemplate=full_parameters("particle_picking_template"),
                particleDiameter=full_parameters("particle_diameter"),
                numberOfParticles=full_parameters("number_of_particles"),
                summaryImageFullPath=full_parameters("summary_image_full_path"),
            )
            session.add(values)
            session.commit()
            self.log.info(
                f"Created ParticlePicker record {values.particlePickerId} "
                f"for DCID {dcid}"
            )
            return {"success": True, "return_value": values.particlePickerId}
        except sqlalchemy.exc.SQLAlchemyError as e:
            self.log.error(
                "Inserting Particle Picker entry caused exception '%s'.",
                e,
                exc_info=True,
            )
            return False

    def do_insert_particle_classification(
        self, parameters, session, message=None, **kwargs
    ):
        if message is None:
            message = {}
        dcid = parameters("dcid")

        def full_parameters(param):
            return message.get(param) or parameters(param)

        try:
            values = models.ParticleClassification(
                particleClassificationId=full_parameters("particle_classification_id"),
                particleClassificationGroupId=full_parameters(
                    "particle_classification_group_id"
                ),
                classNumber=full_parameters("class_number"),
                classImageFullPath=full_parameters("class_image_full_path"),
                particlesPerClass=full_parameters("particles_per_class"),
                rotationAccuracy=full_parameters("rotation_accuracy"),
                translationAccuracy=full_parameters("translation_accuracy"),
                estimatedResolution=full_parameters("estimated_resolution"),
                overallFourierCompleteness=full_parameters(
                    "overall_fourier_completeness"
                ),
                classDistribution=full_parameters("class_distribution"),
                selected=full_parameters("selected"),
            )
            particle_classification = (
                session.query(models.ParticleClassification)
                .filter(
                    models.ParticleClassification.particleClassificationId
                    == values.particleClassificationId,
                )
                .first()
            )
            if particle_classification:
                session.query(models.ParticleClassification).filter(
                    models.ParticleClassification.particleClassificationId
                    == values.particleClassificationId,
                ).update(
                    {
                        k: v
                        for k, v in values.__dict__.items()
                        if k not in ["_sa_instance_state", "particleClassificationId"]
                        and v is not None
                    }
                )
            else:
                session.add(values)
            session.commit()
            self.log.info(
                "Created ParticleClassification record "
                f"{values.particleClassificationId} for DCID {dcid}"
            )
            return {"success": True, "return_value": values.particleClassificationId}
        except sqlalchemy.exc.SQLAlchemyError as e:
            self.log.error(
                "Inserting particle classification entry caused exception '%s'.",
                e,
                exc_info=True,
            )
            return False

    def do_insert_particle_classification_group(
        self, parameters, session, message=None, **kwargs
    ):
        if message is None:
            message = {}
        dcid = parameters("dcid")

        def full_parameters(param):
            return message.get(param) or parameters(param)

        self.log.info(f"Inserting particle classification parameters. DCID: {dcid}")
        try:
            values = models.ParticleClassificationGroup(
                particleClassificationGroupId=full_parameters(
                    "particle_classification_group_id"
                ),
                particlePickerId=full_parameters("particle_picker_id"),
                programId=full_parameters("program_id"),
                type=full_parameters("type"),
                batchNumber=full_parameters("batch_number"),
                numberOfParticlesPerBatch=full_parameters(
                    "number_of_particles_per_batch"
                ),
                numberOfClassesPerBatch=full_parameters("number_of_classes_per_batch"),
                symmetry=full_parameters("symmetry"),
            )
            particle_classification_group = (
                session.query(models.ParticleClassificationGroup)
                .filter(
                    models.ParticleClassificationGroup.particleClassificationGroupId
                    == values.particleClassificationGroupId,
                )
                .first()
            )
            if particle_classification_group:
                session.query(models.ParticleClassificationGroup).filter(
                    models.ParticleClassificationGroup.particleClassificationGroupId
                    == values.particleClassificationGroupId,
                ).update(
                    {
                        k: v
                        for k, v in values.__dict__.items()
                        if k
                        not in ["_sa_instance_state", "particleClassificationGroupId"]
                        and v is not None
                    }
                )
            else:
                session.add(values)
            session.commit()
            self.log.info(
                "Created particle classification group record "
                f"{values.particleClassificationGroupId} for DCID {dcid}"
            )
            return {
                "success": True,
                "return_value": values.particleClassificationGroupId,
            }
        except sqlalchemy.exc.SQLAlchemyError as e:
            self.log.error(
                "Inserting particle classification group entry caused exception '%s'.",
                e,
                exc_info=True,
            )
            return False

    def do_insert_cryoem_initial_model(
        self, parameters, session, message=None, **kwargs
    ):
        if message is None:
            message = {}
        dcid = parameters("dcid")

        def full_parameters(param):
            return message.get(param) or parameters(param)

        self.log.info(f"Inserting CryoEM Initial Model parameters. DCID: {dcid}")
        try:
            if not full_parameters("cryoem_initial_model_id"):
                values_im = models.CryoemInitialModel(
                    resolution=full_parameters("resolution"),
                    numberOfParticles=full_parameters("number_of_particles"),
                )
                session.add(values_im)
                session.commit()
                self.log.info(
                    "Created CryoEM Initial Model record "
                    f"{values_im.cryoemInitialModelId} for DCID {dcid}"
                )
                initial_model_id = values_im.cryoemInitialModelId
            else:
                initial_model_id = full_parameters("cryoem_initial_model_id")
            session.execute(
                models.t_ParticleClassification_has_CryoemInitialModel.insert().values(
                    cryoemInitialModelId=initial_model_id,
                    particleClassificationId=full_parameters(
                        "particle_classification_id"
                    ),
                )
            )
            session.commit()
            return {"success": True, "return_value": initial_model_id}
        except ispyb.ISPyBException as e:
            self.log.error(
                "Inserting CryoEM Initial Model entry caused exception '%s'.",
                e,
                exc_info=True,
            )
            return False

    def do_insert_tomogram(self, parameters, session, message=None, **kwargs):
        if message is None:
            message = {}
        dcid = parameters("dcid")
        self.log.info(f"Inserting Tomogram parameters. DCID: {dcid}")

        def full_parameters(param):
            return message.get(param) or parameters(param)

        try:
            values = models.Tomogram(
                dataCollectionId=full_parameters("dcid"),
                autoProcProgramId=full_parameters("program_id"),
                volumeFile=full_parameters("volume_file"),
                stackFile=full_parameters("stack_file"),
                sizeX=full_parameters("size_x"),
                sizeY=full_parameters("size_y"),
                sizeZ=full_parameters("size_z"),
                pixelSpacing=full_parameters("pixel_spacing"),
                residualErrorMean=full_parameters("residual_error_mean"),
                residualErrorSD=full_parameters("residual_error_sd"),
                xAxisCorrection=full_parameters("x_axis_correction"),
                tiltAngleOffset=full_parameters("tilt_angle_offset"),
                zShift=full_parameters("z_shift"),
                fileDirectory=full_parameters("file_directory"),
                centralSliceImage=full_parameters("central_slice_image"),
                tomogramMovie=full_parameters("tomogram_movie"),
                xyShiftPlot=full_parameters("xy_shift_plot"),
                projXY=full_parameters("proj_xy"),
                projXZ=full_parameters("proj_xz"),
                globalAlignmentQuality=full_parameters("alignment_quality"),
            )
            session.add(values)
            session.commit()
            return {"success": True, "return_value": values.tomogramId}
        except sqlalchemy.exc.SQLAlchemyError as e:
            self.log.error(
                "Inserting Tomogram entry caused exception '%s'.",
                e,
                exc_info=True,
            )
            return False

    def do_insert_tilt_image_alignment(
        self, parameters, session, message=None, **kwargs
    ):
        if message is None:
            message = {}
        dcid = parameters("dcid")
        self.log.info(f"Inserting Tilt Image Alignment parameters. DCID: {dcid}")

        def full_parameters(param):
            return message.get(param) or parameters(param)

        if full_parameters("movie_id"):
            mvid = full_parameters("movie_id")
        else:
            mvid = self._get_movie_id(full_parameters("path"), dcid, session)

        if not mvid:
            self.log.error("No movie ID for tilt image alignment")
            return False

        try:
            values = models.TiltImageAlignment(
                movieId=mvid,
                tomogramId=full_parameters("tomogram_id"),
                defocusU=full_parameters("defocus_u"),
                defocusV=full_parameters("defocus_v"),
                psdFile=full_parameters("psd_file"),
                resolution=full_parameters("resolution"),
                fitQuality=full_parameters("fit_quality"),
                refinedMagnification=full_parameters("refined_magnification"),
                refinedTiltAngle=full_parameters("refined_tilt_angle"),
                refinedTiltAxis=full_parameters("refined_tilt_axis"),
                residualError=full_parameters("residual_error"),
            )
            session.add(values)
            session.commit()
            return {"success": True, "return_value": values.tomogramId}
        except sqlalchemy.exc.SQLAlchemyError as e:
            self.log.error(
                "Inserting Tilt Image Alignment entry caused exception '%s'.",
                e,
                exc_info=True,
            )
            return False

    # These are needed for the old relion-zocalo wrapper
    def do_add_program_attachment(self, parameters, **kwargs):
        params = self.ispyb.mx_processing.get_program_attachment_params()
        params["parentid"] = parameters("program_id")
        try:
            programid = int(params["parentid"])
        except ValueError:
            programid = None
        if not programid:
            self.log.warning("Encountered invalid program ID '%s'", params["parentid"])
            return False
        params["file_name"] = parameters("file_name", replace_variables=False)
        params["file_path"] = parameters("file_path", replace_variables=False)
        params["importance_rank"] = parameters(
            "importance_rank", replace_variables=False
        )
        fqpn = Path(params["file_path"]) / params["file_name"]

        if not fqpn.is_file():
            self.log.error(
                "Not adding attachment '%s' to data processing: File does not exist",
                str(fqpn),
            )
            return False

        params["file_type"] = str(parameters("file_type")).lower()
        if params["file_type"] not in ("log", "result", "graph"):
            self.log.warning(
                "Attachment type '%s' unknown, defaulting to 'log'", params["file_type"]
            )
            params["file_type"] = "log"

        self.log.debug("Writing program attachment to database: %s", params)

        result = self.ispyb.mx_processing.upsert_program_attachment(
            list(params.values())
        )
        return {"success": True, "return_value": result}

    def do_register_processing(self, parameters, **kwargs):
        program = parameters("program")
        cmdline = parameters("cmdline")
        environment = parameters("environment") or ""
        if isinstance(environment, dict):
            environment = ", ".join(
                f"{key}={value}" for key, value in environment.items()
            )
        environment = environment[: min(255, len(environment))]
        rpid = parameters("rpid")
        if rpid and not rpid.isdigit():
            self.log.error(f"Invalid processing id {rpid}")
            return False
        try:
            result = self.ispyb.mx_processing.upsert_program_ex(
                job_id=rpid,
                name=program,
                command=cmdline,
                environment=environment,
            )
            self.log.info(
                f"Registered new program {program} for processing id {rpid} "
                f"with command line {cmdline} and environment {environment} "
                f"with result {result}."
            )
            return {"success": True, "return_value": result}
        except ispyb.ISPyBException as e:
            self.log.error(
                f"Registering new program {program} for processing id {rpid} "
                f"with command line {cmdline} and environment {environment} "
                f"caused exception {e}.",
                exc_info=True,
            )
            return False

    def do_update_processing_status(self, parameters, **kwargs):
        ppid = parameters("program_id")
        message = parameters("message")
        status = parameters("status")
        try:
            result = self.ispyb.mx_processing.upsert_program_ex(
                program_id=ppid,
                status={"success": 1, "failure": 0}.get(status),
                time_start=parameters("start_time"),
                time_update=parameters("update_time"),
                message=message,
            )
            self.log.info(
                f"Updating program {ppid} with status {message}",
            )
            # result is just ppid
            return {"success": True, "return_value": result}
        except ispyb.ISPyBException as e:
            self.log.error(
                f"Updating program {ppid} status: {message} caused exception {e}.",
                exc_info=True,
            )
            return False



from __future__ import annotations

import datetime
import json
import os
import re
from pathlib import Path
from typing import Optional

import workflows.recipe
from pipeliner.api.api_utils import (
    edit_jobstar,
    job_default_parameters_dict,
    write_default_jobstar,
)
from pipeliner.api.manage_project import PipelinerProject
from pipeliner.data_structure import FAIL_FILE, SUCCESS_FILE
from pipeliner.job_factory import read_job
from pipeliner.project_graph import ProjectGraph
from pydantic import BaseModel, Field, ValidationError
from workflows.services.common_service import CommonService

from cryoemservices.util.spa_output_files import create_output_files
from cryoemservices.util.spa_relion_service_options import (
    RelionServiceOptions,
    generate_service_options,
)

# A dictionary of all the available jobs,
# the folder name they run in, and the names of their inputs in the job star
pipeline_spa_jobs: dict[str, dict] = {
    "relion.import.movies": {"folder": "Import", "input_stars": {}},
    "relion.motioncorr.own": {
        "folder": "MotionCorr",
        "input_stars": {"input_star_mics": "movies.star"},
    },
    "relion.motioncorr.motioncor2": {
        "folder": "MotionCorr",
        "input_stars": {"input_star_mics": "movies.star"},
    },
    "icebreaker.micrograph_analysis.micrographs": {
        "folder": "IceBreaker",
        "input_stars": {"in_mics": "corrected_micrographs.star"},
    },
    "icebreaker.micrograph_analysis.enhancecontrast": {
        "folder": "IceBreaker",
        "input_stars": {"in_mics": "corrected_micrographs.star"},
    },
    "icebreaker.micrograph_analysis.summary": {
        "folder": "IceBreaker",
        "input_stars": {"in_mics": "grouped_micrographs.star"},
    },
    "relion.ctffind.ctffind4": {
        "folder": "CtfFind",
        "input_stars": {"input_star_mics": "corrected_micrographs.star"},
    },
    "cryolo.autopick": {
        "folder": "AutoPick",
        "input_stars": {"input_file": "corrected_micrographs.star"},
    },
    "relion.extract": {
        "folder": "Extract",
        "input_stars": {
            "coords_suffix": "autopick.star",
            "star_mics": "micrographs_ctf.star",
        },
    },
    "relion.select.split": {
        "folder": "Select",
        "input_stars": {"fn_data": "particles.star"},
    },
    "icebreaker.micrograph_analysis.particles": {
        "folder": "IceBreaker",
        "input_stars": {
            "in_mics": "grouped_micrographs.star",
            "in_parts": "particles_split1.star",
        },
    },
    "relion.class2d.em": {
        "folder": "Class2D",
        "input_stars": {"fn_img": "particles_split1.star"},
    },
    "relion.class2d.vdam": {
        "folder": "Class2D",
        "input_stars": {"fn_img": "particles_split1.star"},
    },
    "relion.select.class2dauto": {
        "folder": "Select",
        "input_stars": {"fn_model": "run_it020_optimiser.star"},
    },
    "combine_star_files_job": {
        "folder": "Select",
        "input_stars": {"files_to_process": "particles.star"},
    },
    "relion.initialmodel": {
        "folder": "InitialModel",
        "input_stars": {"fn_img": "particles_split1.star"},
    },
    "relion.class3d": {
        "folder": "Class3D",
        "input_stars": {
            "fn_img": "particles_split1.star",
            "fn_ref": "initial_model.mrc",
        },
    },
}


class NodeCreatorParameters(BaseModel):
    job_type: str
    input_file: str = Field(..., min_length=1)
    output_file: str = Field(..., min_length=1)
    relion_options: RelionServiceOptions
    command: str
    stdout: str
    stderr: str
    success: bool = True
    results: Optional[dict] = None


class NodeCreator(CommonService):
    """
    A service for setting up pipeliner jobs
    """

    # Human readable service name
    _service_name = "NodeCreator"

    # Logger name
    _logger_name = "cryoemservices.services.node_creator"

    def initializing(self):
        """Subscribe to a queue. Received messages must be acknowledged."""
        try:
            queue_name = os.environ["NODE_CREATOR_QUEUE"]
        except KeyError:
            queue_name = "node_creator"
        self.log.info(f"Relion node creator service starting for queue {queue_name}")
        workflows.recipe.wrap_subscribe(
            self._transport,
            queue_name,
            self.node_creator,
            acknowledgement=True,
            log_extender=self.extend_log,
            allow_non_recipe_messages=True,
        )

    def node_creator(self, rw, header: dict, message: dict):
        class MockRW:
            def dummy(self, *args, **kwargs):
                pass

        if not rw:
            if (
                not isinstance(message, dict)
                or not message.get("parameters")
                or not message.get("content")
            ):
                self.log.error("Rejected invalid simple message")
                self._transport.nack(header)
                return

            # Create a wrapper-like object that can be passed to functions
            # as if a recipe wrapper was present.
            rw = MockRW()
            rw.transport = self._transport
            rw.recipe_step = {"parameters": message["parameters"]}
            rw.environment = {"has_recipe_wrapper": False}
            rw.set_default_channel = rw.dummy
            rw.send = rw.dummy
            message = message["content"]

        # Read in and validate the parameters
        try:
            if isinstance(message, dict):
                job_info = NodeCreatorParameters(
                    **{**rw.recipe_step.get("parameters", {}), **message}
                )
            else:
                job_info = NodeCreatorParameters(
                    **{**rw.recipe_step.get("parameters", {})}
                )
        except (ValidationError, TypeError) as e:
            self.log.warning(
                f"Node creator parameter validation failed for message: {message} "
                f"and recipe parameters: {rw.recipe_step.get('parameters', {})} "
                f"with exception: {e}"
            )
            rw.transport.nack(header)
            return

        self.log.info(
            f"Received job {job_info.job_type} with output file {job_info.output_file}"
        )
        start_time = datetime.datetime.now()

        # Find the job directory and make sure we are in the processing directory
        job_dir = Path(re.search(".+/job[0-9]{3}", job_info.output_file)[0])
        project_dir = job_dir.parent.parent
        os.chdir(project_dir)

        if not (project_dir / "default_pipeline.star").exists():
            self.log.info("No existing project found, so creating one")
            PipelinerProject(make_new_project=True)

        if not pipeline_spa_jobs.get(job_info.job_type):
            self.log.error(f"Unknown job type {job_info.job_type}")
            rw.transport.nack(header)
            return

        try:
            # Get the options for this job out of the RelionServiceOptions
            pipeline_options = generate_service_options(
                job_info.relion_options,
                job_info.job_type,
            )
            # Work out the name of the input star file and add this to the job.star
            if job_dir.parent.name != "Import":
                ii = 0
                for label, star in pipeline_spa_jobs[job_info.job_type][
                    "input_stars"
                ].items():
                    input_job_dir = Path(
                        re.search(
                            ".+/job[0-9]{3}/", job_info.input_file.split(":")[ii]
                        )[0]
                    )
                    try:
                        pipeline_options[label] = (
                            input_job_dir.relative_to(project_dir) / star
                        )
                    except ValueError:
                        self.log.warning(
                            f"WARNING: {input_job_dir} is not relative to {project_dir}"
                        )
                        pipeline_options[label] = input_job_dir / star
                    ii += 1
            else:
                pipeline_options["fn_in_raw"] = job_info.input_file

            # If this is a new job we need a job.star
            if not Path(f"{job_info.job_type.replace('.', '_')}_job.star").is_file():
                self.log.info(f"Generating options for new job: {job_info.job_type}")
                write_default_jobstar(job_info.job_type)
                params = job_default_parameters_dict(job_info.job_type)
                params.update(pipeline_options)
                _params = {
                    k: str(v) for k, v in params.items() if not isinstance(v, bool)
                }
                _params.update(
                    {
                        k: ("Yes" if v else "No")
                        for k, v in params.items()
                        if isinstance(v, bool)
                    }
                )
                params = _params
                edit_jobstar(
                    f"{job_info.job_type.replace('.', '_')}_job.star",
                    params,
                    f"{job_info.job_type.replace('.', '_')}_job.star",
                )
        except IndexError:
            self.log.error(f"Unknown job type: {job_info.job_type}")
            rw.transport.nack(header)
            return

        # Copy the job.star file
        (job_dir / "job.star").write_bytes(
            Path(f"{job_info.job_type.replace('.', '_')}_job.star").read_bytes()
        )

        # Mark the job completion status
        for exit_file in job_dir.glob("PIPELINER_JOB_EXIT_*"):
            exit_file.unlink()
        if job_info.success:
            (job_dir / SUCCESS_FILE).touch()
        else:
            (job_dir / FAIL_FILE).touch()

        # Get the files and directories relative to the project if possible
        relative_job_dir = (
            job_dir.relative_to(project_dir)
            if job_dir.is_relative_to(project_dir)
            else job_dir
        )
        relative_input_file = (
            Path(job_info.input_file).relative_to(project_dir)
            if Path(job_info.input_file).is_relative_to(project_dir)
            else job_info.input_file
        )
        relative_output_file = (
            Path(job_info.output_file).relative_to(project_dir)
            if Path(job_info.output_file).is_relative_to(project_dir)
            else job_info.output_file
        )

        # Load this job as a pipeliner job to create the nodes
        pipeliner_job = read_job(f"{job_dir}/job.star")
        pipeliner_job.output_dir = str(relative_job_dir) + "/"
        relion_commands = [[], pipeliner_job.get_final_commands()]
        pipeliner_job.prepare_to_run(ignore_invalid_joboptions=True)

        # Write the log files
        with open(job_dir / "run.out", "w") as f:
            f.write(job_info.stdout)
        with open(job_dir / "run.err", "a") as f:
            if job_info.stderr:
                f.write(f"{job_info.stderr}\n")
        with open(job_dir / "note.txt", "a") as f:
            f.write(f"{job_info.command}\n")

        if job_info.success:
            # Write the output files which Relion produces
            extra_output_nodes = create_output_files(
                job_type=job_info.job_type,
                job_dir=relative_job_dir,
                input_file=relative_input_file,
                output_file=relative_output_file,
                relion_options=job_info.relion_options,
                results=job_info.results,
            )
            if extra_output_nodes:
                # Add any extra nodes if they are not already present
                existing_nodes = []
                for node in pipeliner_job.output_nodes:
                    existing_nodes.append(node.name)
                for node in extra_output_nodes.keys():
                    if f"{relative_job_dir}/{node}" not in existing_nodes:
                        pipeliner_job.add_output_node(
                            node,
                            extra_output_nodes[node][0],
                            extra_output_nodes[node][1],
                        )

            # Save the metadata file
            metadata_dict = pipeliner_job.gather_metadata()
            with open(job_dir / "job_metadata.json", "w") as metadata_file:
                metadata_file.write(json.dumps(metadata_dict))

            # Create the results display for the non-pipeliner job
            if job_info.job_type == "combine_star_files_job":
                results_files = job_dir.glob(".results_display*")
                for results_obj in results_files:
                    results_obj.unlink()

                results_displays = pipeliner_job.create_results_display()
                for results_obj in results_displays:
                    results_obj.write_displayobj_file(outdir=str(job_dir))

        # Create the node and default_pipeline.star files in the project directory
        with ProjectGraph(read_only=False) as project:
            process = project.add_job(
                pipeliner_job,
                as_status=("Succeeded" if job_info.success else "Failed"),
                do_overwrite=True,
            )
            # Add the job commands to the process .CCPEM_pipeliner_jobinfo file
            if not (job_dir / ".CCPEM_pipeliner_jobinfo").exists():
                process.update_jobinfo_file(action="Run", command_list=relion_commands)
            # Generate the default_pipeline.star file
            project.check_process_completion()
            # Copy the default_pipeline.star file
            (job_dir / "default_pipeline.star").write_bytes(
                Path("default_pipeline.star").read_bytes()
            )

        end_time = datetime.datetime.now()
        self.log.info(
            f"Processed outputs from job {job_info.job_type}, "
            f"in {(end_time - start_time).total_seconds()} seconds."
        )
        rw.transport.ack(header)



from __future__ import annotations

from pathlib import Path

from pipeliner.data_structure import SELECT_DIR
from pipeliner.display_tools import mini_montage_from_starfile
from pipeliner.job_options import BooleanJobOption, IntJobOption, StringJobOption
from pipeliner.nodes import NODE_PARTICLESDATA, Node
from pipeliner.pipeliner_job import ExternalProgram, PipelinerJob

try:
    from pipeliner.pipeliner_job import PipelinerCommand
except ImportError:
    PipelinerCommand = None

COMBINE_STAR_NAME = "combine_star_files"


class ProcessStarFiles(PipelinerJob):
    PROCESS_NAME = "combine_star_files_job"
    OUT_DIR = SELECT_DIR

    def __init__(self):
        super().__init__()

        self.jobinfo.display_name = "Particle star file merging"
        self.jobinfo.short_desc = "Combine and split star files of particles"
        self.jobinfo.long_desc = (
            "Combine star files of particles, then optionally split them again."
        )

        self.jobinfo.programs = [ExternalProgram(command=COMBINE_STAR_NAME)]

        self.joboptions["files_to_process"] = StringJobOption(
            label="The star files from which to combine particles",
            is_required=True,
            help_text="The names of the star files, separated by spaces.",
        )
        self.joboptions["do_split"] = BooleanJobOption(
            label="Whether to split the combined star file", default_value=False
        )
        self.joboptions["n_files"] = IntJobOption(
            label="Number of files to split the combined file into",
            default_value=-1,
            help_text=(
                "Provide either the number of files to split into,"
                " or the number of particles per file."
            ),
            deactivate_if=[("do_split", "is", "False")],
        )
        self.joboptions["split_size"] = IntJobOption(
            label="Number of particles to put in each split",
            default_value=-1,
            help_text=(
                "Provide either the number of files to split into,"
                " or the number of particles per file."
            ),
            deactivate_if=[("do_split", "is", "False")],
        )

        self.set_joboption_order(
            ["files_to_process", "do_split", "n_files", "split_size"]
        )

    def get_commands(self):
        """Construct the command for combining and splitting star files"""
        command = [COMBINE_STAR_NAME]
        file_list = self.joboptions["files_to_process"].get_string().split(" ")
        command.extend(file_list)
        command.extend(["--output_dir", str(self.output_dir)])

        if self.joboptions["do_split"].get_boolean():
            command.extend(["--split"])

            if (
                self.joboptions["n_files"].get_number() <= 0
                and self.joboptions["split_size"].get_number() <= 0
            ):
                raise ValueError(
                    "ERROR: When splitting the combined STAR file into subsets,"
                    " set n_files or split_size to a positive value"
                )

            if self.joboptions["n_files"].get_number() > 0:
                command.extend(["--n_files", self.joboptions["n_files"].get_string()])

            if self.joboptions["split_size"].get_number() > 0:
                command.extend(
                    ["--split_size", self.joboptions["split_size"].get_string()]
                )

        # Add files as input nodes, as long as they are not also the output node
        for particle_file in file_list:
            if particle_file != str(Path(self.output_dir) / "particles_all.star"):
                self.input_nodes.append(Node(particle_file, NODE_PARTICLESDATA))
        self.output_nodes.append(
            Node(str(Path(self.output_dir) / "particles_all.star"), NODE_PARTICLESDATA)
        )

        if PipelinerCommand is None:
            return [command]
        pipeliner_commands = [PipelinerCommand([command], relion_control=False)]
        return pipeliner_commands

    def create_output_nodes(self):
        self.add_output_node("particles_all.star", NODE_PARTICLESDATA, ["relion"])

    def post_run_actions(self):
        """Find any output files produced by the splitting"""
        output_files = Path(self.output_dir).glob("particles_split*.star")
        for split in output_files:
            self.output_nodes.append(Node(str(split), NODE_PARTICLESDATA))

    def create_results_display(self):
        with open(
            Path(self.output_dir) / "class_averages.star", "r"
        ) as all_classes, open(
            Path(self.output_dir) / ".class_display_tmp.star", "w"
        ) as display_classes:
            for line in range(200):
                class_line = all_classes.readline()
                if not class_line:
                    break
                display_classes.write(class_line)
        with open(
            Path(self.output_dir) / "particles_all.star", "r"
        ) as all_particles, open(
            Path(self.output_dir) / ".particles_display_tmp.star", "w"
        ) as display_particles:
            for line in range(200):
                particles_line = all_particles.readline()
                if not particles_line:
                    break
                display_particles.write(particles_line)
        output_dobs = [
            mini_montage_from_starfile(
                starfile=str(Path(self.output_dir) / ".class_display_tmp.star"),
                block="",
                column="_rlnReferenceImage",
                outputdir=self.output_dir,
                nimg=100,
                title="Selected 2D classes",
            ),
            mini_montage_from_starfile(
                starfile=str(Path(self.output_dir) / ".particles_display_tmp.star"),
                block="particles",
                column="_rlnImageName",
                outputdir=self.output_dir,
                nimg=20,
                title="Examples of selected particles",
            ),
        ]
        (Path(self.output_dir) / ".class_display_tmp.star").unlink()
        (Path(self.output_dir) / ".particles_display_tmp.star").unlink()
        return output_dobs



from __future__ import annotations

import re
import subprocess
from pathlib import Path

import numpy as np
import workflows.recipe
from gemmi import cif
from pydantic import BaseModel, Field, ValidationError
from workflows.services.common_service import CommonService

from cryoemservices.util.spa_relion_service_options import RelionServiceOptions


class CryoloParameters(BaseModel):
    input_path: str = Field(..., min_length=1)
    output_path: str = Field(..., min_length=1)
    pixel_size: float
    cryolo_config_file: str = "/dls_sw/apps/EM/crYOLO/phosaurus_models/config.json"
    cryolo_model_weights: str = (
        "/dls_sw/apps/EM/crYOLO/phosaurus_models/gmodel_phosnet_202005_N63_c17.h5"
    )
    cryolo_threshold: float = 0.3
    retained_fraction: float = 1
    min_particles: int = 30
    cryolo_command: str = "cryolo_predict.py"
    particle_diameter: float = None
    on_the_fly: bool = True
    mc_uuid: int
    picker_uuid: int
    relion_options: RelionServiceOptions
    ctf_values: dict = {}


class CrYOLO(CommonService):
    """
    A service that runs crYOLO particle picking
    """

    # Human readable service name
    _service_name = "CrYOLO"

    # Logger name
    _logger_name = "cryoemservices.services.cryolo"

    # Job name
    job_type = "cryolo.autopick"

    # Values to extract for ISPyB
    number_of_particles: int

    def initializing(self):
        """Subscribe to a queue. Received messages must be acknowledged."""
        self.log.info("crYOLO service starting")
        workflows.recipe.wrap_subscribe(
            self._transport,
            "cryolo",
            self.cryolo,
            acknowledgement=True,
            log_extender=self.extend_log,
            allow_non_recipe_messages=True,
        )

    def parse_cryolo_output(self, cryolo_stdout: str):
        """
        Read the output logs of cryolo to determine
        the number of particles that are picked
        """
        for line in cryolo_stdout.split("\n"):
            if "particles in total has been found" in line:
                line_split = line.split()
                self.number_of_particles += int(line_split[0])

            if line.startswith("Deleted"):
                line_split = line.split()
                self.number_of_particles -= int(line_split[1])

    def cryolo(self, rw, header: dict, message: dict):
        """
        Main function which interprets received messages, runs cryolo
        and sends messages to the ispyb and image services
        """

        class MockRW:
            def dummy(self, *args, **kwargs):
                pass

        if not rw:
            if (
                not isinstance(message, dict)
                or not message.get("parameters")
                or not message.get("content")
            ):
                self.log.error("Rejected invalid simple message")
                self._transport.nack(header)
                return

            # Create a wrapper-like object that can be passed to functions
            # as if a recipe wrapper was present.
            rw = MockRW()
            rw.transport = self._transport
            rw.recipe_step = {"parameters": message["parameters"]}
            rw.environment = {"has_recipe_wrapper": False}
            rw.set_default_channel = rw.dummy
            rw.send = rw.dummy
            message = message["content"]

        # Reset number of particles
        self.number_of_particles = 0

        try:
            if isinstance(message, dict):
                cryolo_params = CryoloParameters(
                    **{**rw.recipe_step.get("parameters", {}), **message}
                )
            else:
                cryolo_params = CryoloParameters(
                    **{**rw.recipe_step.get("parameters", {})}
                )
        except (ValidationError, TypeError) as e:
            self.log.warning(
                f"crYOLO parameter validation failed for message: {message} "
                f"and recipe parameters: {rw.recipe_step.get('parameters', {})} "
                f"with exception: {e}"
            )
            rw.transport.nack(header)
            return

        # Check if this file has been run before
        if Path(cryolo_params.output_path).is_file():
            job_is_rerun = True
        else:
            job_is_rerun = False

        # CrYOLO requires running in the project directory or job directory
        job_dir = Path(re.search(".+/job[0-9]{3}/", cryolo_params.output_path)[0])
        job_dir.mkdir(parents=True, exist_ok=True)

        # Construct a command to run cryolo with the given parameters
        command = cryolo_params.cryolo_command.split()
        command.extend((["--conf", cryolo_params.cryolo_config_file]))
        command.extend((["-o", str(job_dir)]))
        if cryolo_params.on_the_fly:
            command.extend((["--otf"]))

        cryolo_flags = {
            "cryolo_model_weights": "--weights",
            "input_path": "-i",
            "cryolo_threshold": "--threshold",
            "cryolo_gpus": "--gpu",
        }

        for k, v in cryolo_params.dict().items():
            if v and (k in cryolo_flags):
                command.extend((cryolo_flags[k], str(v)))

        self.log.info(
            f"Input: {cryolo_params.input_path} "
            + f"Output: {cryolo_params.output_path}"
        )

        # Run cryolo
        result = subprocess.run(command, cwd=job_dir, capture_output=True)

        # Remove cryosparc file to minimise crashes
        (job_dir / "CRYOSPARC/cryosparc.star").unlink(missing_ok=True)

        # Read in the stdout from cryolo
        self.parse_cryolo_output(result.stdout.decode("utf8", "replace"))

        # Read in the cbox file for particle selection and finding sizes
        try:
            cbox_file = cif.read_file(
                str(
                    job_dir
                    / f"CBOX/{Path(cryolo_params.output_path).with_suffix('.cbox').name}"
                )
            )
            cbox_block = cbox_file.find_block("cryolo")
            cbox_sizes = np.append(
                np.array(cbox_block.find_loop("_EstWidth"), dtype=float),
                np.array(cbox_block.find_loop("_EstHeight"), dtype=float),
            )
            cbox_confidence = np.append(
                np.array(cbox_block.find_loop("_Confidence"), dtype=float),
                np.array(cbox_block.find_loop("_Confidence"), dtype=float),
            )

            # Select only a fraction of particles based on confidence if requested
            if (
                cryolo_params.retained_fraction < 1
                and cryolo_params.retained_fraction > 0
            ):
                particles_confidence = np.array(
                    cbox_block.find_loop("_Confidence"), dtype=float
                )
                if len(particles_confidence) <= cryolo_params.min_particles:
                    cryolo_threshold = 0.1
                elif (
                    len(particles_confidence) * cryolo_params.retained_fraction
                    < cryolo_params.min_particles
                ):
                    cryolo_threshold = sorted(particles_confidence, reverse=True)[
                        cryolo_params.min_particles
                    ]
                else:
                    cryolo_threshold = np.quantile(
                        particles_confidence, 1 - cryolo_params.retained_fraction
                    )

                self.log.info(
                    f"Selecting particles with confidence above {cryolo_threshold}"
                )

                particles_x_all = np.array(
                    cbox_block.find_loop("_CoordinateX"), dtype=float
                )
                particles_y_all = np.array(
                    cbox_block.find_loop("_CoordinateY"), dtype=float
                )
                box_size_x_all = np.array(cbox_block.find_loop("_Width"), dtype=float)
                box_size_y_all = np.array(cbox_block.find_loop("_Height"), dtype=float)

                thresholded_x = (particles_x_all + box_size_x_all / 2)[
                    particles_confidence > cryolo_threshold
                ]
                thresholded_y = (particles_y_all + box_size_y_all / 2)[
                    particles_confidence > cryolo_threshold
                ]

                # Update the number of particles
                self.number_of_particles = len(thresholded_x)

                # Rewrite the star file with only the selected particles
                with open(job_dir / cryolo_params.output_path, "w") as particles_star:
                    particles_star.write(
                        "\ndata_\n\nloop_\n_rlnCoordinateX #1\n_rlnCoordinateY #2\n"
                    )
                    for particle in range(len(thresholded_x)):
                        particles_star.write(
                            f"{thresholded_x[particle]} {thresholded_y[particle]}\n"
                        )
            else:
                cryolo_threshold = cryolo_params.cryolo_threshold
            cryolo_params.relion_options.cryolo_threshold = cryolo_threshold

            # Get the diameters of the particles in Angstroms for Murfey
            cryolo_particle_sizes = (
                cbox_sizes[cbox_confidence > cryolo_threshold]
                * cryolo_params.pixel_size
            )
        except (FileNotFoundError, OSError, AttributeError):
            cryolo_particle_sizes = []

        # Register the cryolo job with the node creator
        self.log.info(f"Sending {self.job_type} to node creator")
        node_creator_parameters = {
            "job_type": self.job_type,
            "input_file": cryolo_params.input_path,
            "output_file": cryolo_params.output_path,
            "relion_options": dict(cryolo_params.relion_options),
            "command": " ".join(command),
            "stdout": result.stdout.decode("utf8", "replace"),
            "stderr": result.stderr.decode("utf8", "replace"),
        }
        if (
            result.returncode
            and result.stderr.decode("utf8", "replace").split("\n")[-1]
            == "IndexError: list index out of range"
        ):
            # If Cryolo failed because there are no picks then consider it a success
            result.returncode = 0

        if result.returncode:
            node_creator_parameters["success"] = False
        else:
            node_creator_parameters["success"] = True
        if not job_is_rerun:
            # Only do the node creator inserts for new files
            if isinstance(rw, MockRW):
                rw.transport.send(
                    destination="node_creator",
                    message={"parameters": node_creator_parameters, "content": "dummy"},
                )
            else:
                rw.send_to("node_creator", node_creator_parameters)

        # End here if the command failed
        if result.returncode:
            self.log.error(
                f"crYOLO failed with exitcode {result.returncode}:\n"
                + result.stderr.decode("utf8", "replace")
            )
            rw.transport.nack(header)
            return

        # Forward results to ISPyB
        ispyb_parameters = {
            "ispyb_command": "buffer",
            "buffer_lookup": {"motion_correction_id": cryolo_params.mc_uuid},
            "buffer_command": {"ispyb_command": "insert_particle_picker"},
            "buffer_store": cryolo_params.picker_uuid,
            "particle_picking_template": cryolo_params.cryolo_model_weights,
            "number_of_particles": self.number_of_particles,
            "summary_image_full_path": str(
                Path(cryolo_params.output_path).with_suffix(".jpeg")
            ),
        }
        if cryolo_params.particle_diameter:
            ispyb_parameters["particle_diameter"] = cryolo_params.particle_diameter
        self.log.info(f"Sending to ispyb {ispyb_parameters}")
        if isinstance(rw, MockRW):
            rw.transport.send(
                destination="ispyb_connector",
                message={
                    "parameters": ispyb_parameters,
                    "content": {"dummy": "dummy"},
                },
            )
        else:
            rw.send_to("ispyb_connector", ispyb_parameters)

        # Extract results for images service
        try:
            with open(cryolo_params.output_path, "r") as coords_file:
                coords = [line.split() for line in coords_file][6:]
        except FileNotFoundError:
            coords = []

        # Forward results to images service
        self.log.info("Sending to images service")
        if isinstance(rw, MockRW):
            rw.transport.send(
                destination="images",
                message={
                    "image_command": "picked_particles",
                    "file": cryolo_params.input_path,
                    "coordinates": coords,
                    "pixel_size": cryolo_params.pixel_size,
                    "diameter": (
                        cryolo_params.particle_diameter
                        if cryolo_params.particle_diameter
                        else 160
                    ),
                    "outfile": str(
                        Path(cryolo_params.output_path).with_suffix(".jpeg")
                    ),
                },
            )
        else:
            rw.send_to(
                "images",
                {
                    "image_command": "picked_particles",
                    "file": cryolo_params.input_path,
                    "coordinates": coords,
                    "pixel_size": cryolo_params.pixel_size,
                    "diameter": (
                        cryolo_params.particle_diameter
                        if cryolo_params.particle_diameter
                        else 160
                    ),
                    "outfile": str(
                        Path(cryolo_params.output_path).with_suffix(".jpeg")
                    ),
                },
            )

        # Gather results needed for particle extraction
        extraction_params = {
            "ctf_values": cryolo_params.ctf_values,
            "micrographs_file": cryolo_params.input_path,
            "coord_list_file": cryolo_params.output_path,
        }
        job_number = int(re.search("/job[0-9]{3}/", cryolo_params.output_path)[0][4:7])
        extraction_params["extract_file"] = str(
            Path(
                re.sub(
                    "MotionCorr/job002/.+",
                    f"Extract/job{job_number + 1:03}/Movies/",
                    cryolo_params.input_path,
                )
            )
            / (Path(cryolo_params.input_path).stem + "_extract.star")
        )

        # Forward results to murfey
        self.log.info("Sending to Murfey for particle extraction")
        if isinstance(rw, MockRW):
            rw.transport.send(
                destination="murfey_feedback",
                message={
                    "register": "picked_particles",
                    "motion_correction_id": cryolo_params.mc_uuid,
                    "micrograph": cryolo_params.input_path,
                    "particle_diameters": list(cryolo_particle_sizes),
                    "extraction_parameters": extraction_params,
                },
            )
        else:
            rw.send_to(
                "murfey_feedback",
                {
                    "register": "picked_particles",
                    "motion_correction_id": cryolo_params.mc_uuid,
                    "micrograph": cryolo_params.input_path,
                    "particle_diameters": list(cryolo_particle_sizes),
                    "extraction_parameters": extraction_params,
                },
            )

        self.log.info(f"Done {self.job_type} for {cryolo_params.input_path}.")
        rw.transport.ack(header)
